@@ -9,12 +9,12 @@ use futures::TryFutureExt;
 use log::debug;
 
 use tc_transact::fs::BlockData;
-use tc_transact::lock::Mutate;
 use tc_transact::TxnId;
+use tc_value::Value;
 use tcgeneric::{TCPathBuf, Tuple};
 
 use crate::chain::{BLOCK_SIZE, EXT};
-use crate::scalar::{Scalar, Value};
+use crate::scalar::Scalar;
 
 #[derive(Clone, Eq, PartialEq)]
 pub enum Mutation {
@@ -177,19 +177,6 @@ impl ChainBlock {
     }
 }
 
-#[async_trait]
-impl Mutate for ChainBlock {
-    type Pending = Self;
-
-    fn diverge(&self, _txn_id: &TxnId) -> Self::Pending {
-        self.clone()
-    }
-
-    async fn converge(&mut self, new_value: Self::Pending) {
-        *self = new_value;
-    }
-}
-
 #[async_trait]
 impl BlockData for ChainBlock {
     fn ext() -> &'static str {
