@@ -16,17 +16,18 @@ use tc_table::TableInstance;
 #[cfg(feature = "tensor")]
 use tc_tensor::TensorAccess;
 use tc_transact::fs::*;
-use tc_transact::lock::{Mutable, TxnLock};
+use tc_transact::lock::TxnLock;
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
+use tc_value::Value;
 use tcgeneric::{
-    label, Id, Instance, Label, Map, NativeClass, TCPathBuf, TCStream, TCTryStream, Tuple,
+    label, Id, Instance, Label, Map, NativeClass, TCBoxStream, TCBoxTryStream, TCPathBuf, Tuple,
 };
 
 use crate::chain::{ChainType, Subject, CHAIN, NULL_HASH};
 use crate::collection::*;
 use crate::fs;
 use crate::route::Public;
-use crate::scalar::{OpRef, Scalar, TCRef, Value};
+use crate::scalar::{OpRef, Scalar, TCRef};
 use crate::state::{State, StateView};
 use crate::txn::Txn;
 
@@ -38,12 +39,12 @@ const DATA: Label = label("data");
 pub struct History {
     dir: fs::Dir,
     file: fs::File<ChainBlock>,
-    latest: TxnLock<Mutable<u64>>,
+    latest: TxnLock<u64>,
 }
 
 impl History {
     fn new(latest: u64, dir: fs::Dir, file: fs::File<ChainBlock>) -> Self {
-        let latest = TxnLock::new("latest block ordinal", latest.into());
+        let latest = TxnLock::new("latest block ordinal", latest);
         Self { dir, latest, file }
     }
 
@@ -66,7 +67,7 @@ impl History {
 
     pub async fn append_put(
         &self,
-        txn: Txn,
+        txn: &Txn,
         path: TCPathBuf,
         key: Value,
         value: State,
@@ -85,7 +86,7 @@ impl History {
         Ok(())
     }
 
-    async fn save_state(&self, txn: Txn, state: State) -> TCResult<Scalar> {
+    async fn save_state(&self, txn: &Txn, state: State) -> TCResult<Scalar> {
         if state.is_ref() {
             return Err(TCError::bad_request(
                 "cannot update Chain with reference: {}",
@@ -257,8 +258,14 @@ impl History {
         if let Some((last_txn_id, ops)) = last_block.mutations().iter().last() {
             for op in ops {
                 let result = match op {
-                    Mutation::Delete(path, key) => subject.delete(txn, path, key.clone()).await,
+                    Mutation::Delete(path, key) => {
+                        debug!("replay DELETE {}{}: {}", subject, path, key);
+
+                        subject.delete(txn, path, key.clone()).await
+                    }
                     Mutation::Put(path, key, value) => {
+                        debug!("replay PUT {}{}: {} <- {}", subject, path, key, value);
+
                         self.resolve(txn, value.clone())
                             .and_then(|value| subject.put(txn, path, key.clone(), value))
                             .await
@@ -323,7 +330,7 @@ impl History {
                         }
                         Mutation::Put(path, key, value) => {
                             let value = other.resolve(txn, value).await?;
-                            let value_ref = self.save_state(txn.clone(), value.clone()).await?;
+                            let value_ref = self.save_state(txn, value.clone()).await?;
 
                             if append {
                                 dest.append_put(*past_txn_id, path.clone(), key.clone(), value_ref);
@@ -596,7 +603,7 @@ impl de::Visitor for HistoryVisitor {
                 ));
             }
 
-            let mutations = parse_block_state(&history, txn.clone(), block_data)
+            let mutations = parse_block_state(&history, &txn, block_data)
                 .map_err(de::Error::custom)
                 .await?;
 
@@ -630,7 +637,7 @@ impl de::Visitor for HistoryVisitor {
                 ));
             }
 
-            let mutations = parse_block_state(&history, txn.clone(), block_data)
+            let mutations = parse_block_state(&history, &txn, block_data)
                 .map_err(de::Error::custom)
                 .await?;
 
@@ -645,7 +652,7 @@ impl de::Visitor for HistoryVisitor {
 
 async fn parse_block_state(
     history: &History,
-    txn: Txn,
+    txn: &Txn,
     block_data: Map<Tuple<State>>,
 ) -> TCResult<BTreeMap<TxnId, Vec<Mutation>>> {
     let mut mutations = BTreeMap::new();
@@ -661,7 +668,7 @@ async fn parse_block_state(
                 parsed.push(Mutation::Delete(path, key));
             } else if op.matches::<(TCPathBuf, Value, State)>() {
                 let (path, key, value) = op.opt_cast_into().unwrap();
-                let value = history.save_state(txn.clone(), value).await?;
+                let value = history.save_state(txn, value).await?;
                 parsed.push(Mutation::Put(path, key, value));
             } else {
                 return Err(TCError::bad_request(
@@ -678,7 +685,7 @@ async fn parse_block_state(
 }
 
 pub type HistoryView<'en> =
-    en::SeqStream<TCResult<HistoryBlockView<'en>>, TCTryStream<'en, HistoryBlockView<'en>>>;
+    en::SeqStream<TCResult<HistoryBlockView<'en>>, TCBoxTryStream<'en, HistoryBlockView<'en>>>;
 
 #[async_trait]
 impl<'en> IntoView<'en, fs::Dir> for History {
@@ -709,16 +716,16 @@ impl<'en> IntoView<'en, fs::Dir> for History {
                         let mutations = stream::iter(mutations)
                             .then(move |op| Box::pin(load_history(this.clone(), op, txn.clone())));
 
-                        let mutations: TCTryStream<'en, MutationView<'en>> = Box::pin(mutations);
+                        let mutations: TCBoxTryStream<'en, MutationView<'en>> = Box::pin(mutations);
                         let mutations = en::SeqStream::from(mutations);
                         (past_txn_id, Ok(mutations))
                     });
 
-                let map: TCStream<'en, (TxnId, TCResult<MutationViewSeq<'en>>)> = Box::pin(map);
+                let map: TCBoxStream<'en, (TxnId, TCResult<MutationViewSeq<'en>>)> = Box::pin(map);
                 (block.last_hash().clone(), en::MapStream::from(map))
             });
 
-        let seq: TCTryStream<'en, HistoryBlockView<'en>> = Box::pin(seq);
+        let seq: TCBoxTryStream<'en, HistoryBlockView<'en>> = Box::pin(seq);
         Ok(en::SeqStream::from(seq))
     }
 }
@@ -749,14 +756,14 @@ async fn load_history<'a>(history: History, op: Mutation, txn: Txn) -> TCResult<
 }
 
 type MutationViewSeq<'en> =
-    en::SeqStream<TCResult<MutationView<'en>>, TCTryStream<'en, MutationView<'en>>>;
+    en::SeqStream<TCResult<MutationView<'en>>, TCBoxTryStream<'en, MutationView<'en>>>;
 
 type HistoryBlockView<'en> = (
     Bytes,
     en::MapStream<
         TxnId,
         TCResult<MutationViewSeq<'en>>,
-        TCStream<'en, (TxnId, TCResult<MutationViewSeq<'en>>)>,
+        TCBoxStream<'en, (TxnId, TCResult<MutationViewSeq<'en>>)>,
     >,
 );
 
