@@ -8,12 +8,13 @@ use uplock::RwLock;
 use tc_error::*;
 use tc_transact::lock::TxnLock;
 use tc_transact::Transaction;
+use tc_value::{Link, LinkHost, Value};
 use tcgeneric::*;
 
 use crate::chain::{self, Chain, ChainType, Schema};
 use crate::fs;
 use crate::object::{InstanceClass, InstanceExt};
-use crate::scalar::{Link, LinkHost, OpRef, Refer, Scalar, Value};
+use crate::scalar::{OpRef, Refer, Scalar};
 use crate::txn::{Actor, Txn, TxnId};
 
 use super::Cluster;
@@ -79,16 +80,6 @@ pub async fn instantiate(
                     op_def.dereference_self(link.path())
                 };
 
-                for (id, provider) in op_def.form() {
-                    // make sure all writes to a chain subject are recorded
-                    if provider.is_derived_write() {
-                        return Err(TCError::unsupported(format!(
-                            "write op {} may not write to a derived view: {}",
-                            id, provider
-                        )));
-                    }
-                }
-
                 cluster_proto.insert(id, Scalar::Op(op_def));
             }
             other => {
@@ -123,11 +114,8 @@ pub async fn instantiate(
         classes,
         confirmed: RwLock::new(txn_id),
         owned: RwLock::new(HashMap::new()),
-        installed: TxnLock::new(
-            format!("Cluster {} installed deps", link),
-            HashMap::new().into(),
-        ),
-        replicas: TxnLock::new(format!("Cluster {} replicas", link), replicas.into()),
+        installed: TxnLock::new(format!("Cluster {} installed deps", link), HashMap::new()),
+        replicas: TxnLock::new(format!("Cluster {} replicas", link), replicas),
     };
 
     let class = InstanceClass::new(Some(link), cluster_proto.into());
