@@ -18,13 +18,14 @@ use tc_error::*;
 #[cfg(feature = "tensor")]
 use tc_tensor::TensorType;
 use tc_transact::fs::{self, BlockData};
-use tc_transact::lock::{Mutable, TxnLock};
+use tc_transact::lock::TxnLock;
 use tc_transact::{Transact, TxnId};
+use tc_value::Value;
 use tcgeneric::{Id, PathSegment};
 
 use crate::chain::{self, ChainBlock};
 use crate::collection::CollectionType;
-use crate::scalar::{ScalarType, Value};
+use crate::scalar::ScalarType;
 use crate::state::StateType;
 
 use super::{dir_contents, file_ext, file_name, fs_path, Cache, DirContents, File};
@@ -203,7 +204,7 @@ impl fmt::Display for DirEntry {
 pub struct Dir {
     path: PathBuf,
     cache: Cache,
-    contents: TxnLock<Mutable<HashMap<PathSegment, DirEntry>>>,
+    contents: TxnLock<HashMap<PathSegment, DirEntry>>,
 }
 
 impl Dir {
@@ -246,7 +247,7 @@ impl Dir {
                 Ok(Dir {
                     path,
                     cache,
-                    contents: TxnLock::new(lock_name, contents.into()),
+                    contents: TxnLock::new(lock_name, contents),
                 })
             } else {
                 Err(TCError::internal(format!(
@@ -317,7 +318,7 @@ impl fs::Dir for Dir {
             cache: self.cache.clone(),
             contents: TxnLock::new(
                 format!("transactional subdirectory at {}", name),
-                HashMap::new().into(),
+                HashMap::new(),
             ),
         };
 
