@@ -10,12 +10,12 @@ use log::debug;
 use safecast::*;
 
 use tc_error::*;
+use tc_value::{Link, Value};
 use tcgeneric::*;
 
 use crate::cluster::Cluster;
-use crate::collection::CollectionType;
 use crate::object::InstanceExt;
-use crate::route::{Public, Route};
+use crate::route::{Public, Static};
 use crate::scalar::*;
 use crate::state::*;
 use crate::txn::*;
@@ -54,14 +54,12 @@ impl Kernel {
             if key.is_none() {
                 Ok(Value::from(Bytes::copy_from_slice(self.actor.public_key().as_bytes())).into())
             } else {
-                Err(TCError::method_not_allowed(
-                    OpRefType::Get,
-                    self,
-                    TCPath::from(path),
-                ))
+                Err(TCError::not_found(format!(
+                    "{} at {}",
+                    key,
+                    TCPath::from(path)
+                )))
             }
-        } else if let Some(class) = StateType::from_path(path) {
-            construct_state(txn, class, key).await
         } else if let Some((suffix, cluster)) = self.hosted.get(path) {
             debug!(
                 "GET {}: {} from cluster {}",
@@ -71,33 +69,8 @@ impl Kernel {
             );
 
             cluster.get(&txn, suffix, key).await
-        } else if path.len() > 1 && &path[..2] == &crate::collection::PREFIX[..] {
-            let mut i = path.len() - 1;
-            while i > 0 {
-                if let Some(class) = CollectionType::from_path(&path[..i]) {
-                    if let Some(handler) = class.route(&path[i..]) {
-                        if let Some(handler) = handler.get() {
-                            return handler(txn.clone(), key).await;
-                        }
-                    }
-                }
-
-                i -= 1;
-            }
-
-            Err(TCError::not_found(TCPath::from(path)))
-        } else if &path[0] == "error" && path.len() == 2 {
-            let message = String::try_cast_from(key, |v| {
-                TCError::bad_request("cannot cast into error message string from", v)
-            })?;
-
-            if let Some(err_type) = error_type(&path[1]) {
-                Err(TCError::new(err_type, message))
-            } else {
-                Err(TCError::not_found(TCPath::from(path)))
-            }
         } else {
-            Err(TCError::not_found(TCPath::from(path)))
+            Static.get(txn, path, key).await
         }
     }
 
@@ -175,7 +148,7 @@ impl Kernel {
             })
             .await
         } else {
-            Err(TCError::not_found(TCPath::from(path)))
+            Static.put(txn, path, key, value).await
         }
     }
 
@@ -198,11 +171,20 @@ impl Kernel {
 
             if Vec::<(Id, State)>::can_cast_from(&data) {
                 let op_def: Vec<(Id, State)> = data.opt_cast_into().unwrap();
-                OpDef::call(op_def, txn, context).await
+                let capture = if let Some((capture, _)) = op_def.last() {
+                    capture.clone()
+                } else {
+                    return Ok(State::default());
+                };
+
+                let executor: Executor<State> = Executor::new(&txn, None, op_def);
+                executor.capture(capture).await
             } else {
-                data.resolve(&ExeScope::new(&State::default(), context), &txn)
-                    .await
+                data.resolve(&ExeScope::new(None, context), &txn).await
             }
+        } else if let Some(class) = StateType::from_path(path) {
+            let params = data.try_into()?;
+            class.post(txn, path, params).await
         } else if let Some((suffix, cluster)) = self.hosted.get(path) {
             let params: Map<State> = data.try_into()?;
 
@@ -224,7 +206,8 @@ impl Kernel {
                 .await
             }
         } else {
-            Err(TCError::not_found(TCPath::from(path)))
+            let params = data.try_into()?;
+            Static.post(txn, path, params).await
         }
     }
 
@@ -241,6 +224,12 @@ impl Kernel {
                     TCPath::from(path),
                 ))
             }
+        } else if let Some(class) = StateType::from_path(path) {
+            Err(TCError::method_not_allowed(
+                OpRefType::Post,
+                class,
+                TCPath::default(),
+            ))
         } else if let Some((suffix, cluster)) = self.hosted.get(path) {
             if suffix.is_empty() && key.is_none() {
                 // it's a rollback message
@@ -294,18 +283,8 @@ impl Kernel {
                 cluster.replicate_write(txn.clone(), write).await
             })
             .await
-        } else if &path[0] == "error" && path.len() == 2 {
-            if let Some(class) = error_type(&path[1]) {
-                Err(TCError::method_not_allowed(
-                    OpRefType::Delete,
-                    class,
-                    TCPath::from(path),
-                ))
-            } else {
-                Err(TCError::not_found(TCPath::from(path)))
-            }
         } else {
-            Err(TCError::not_found(TCPath::from(path)))
+            Static.delete(txn, path, key).await
         }
     }
 }
@@ -350,10 +329,10 @@ fn execute<
 
             if result.is_ok() {
                 debug!("commit {}", cluster);
-                cluster.distribute_commit(txn).await?;
+                cluster.distribute_commit(&txn).await?;
             } else {
                 debug!("rollback {}", cluster);
-                cluster.distribute_rollback(txn).await;
+                cluster.distribute_rollback(&txn).await;
             }
 
             result
@@ -361,63 +340,6 @@ fn execute<
     })
 }
 
-async fn construct_state(txn: &Txn, class: StateType, value: Value) -> TCResult<State> {
-    match class {
-        StateType::Collection(class) => {
-            if let Some(handler) = class.route(&[]) {
-                if let Some(handler) = handler.get() {
-                    return handler(txn.clone(), value).await;
-                }
-            }
-
-            Err(TCError::not_implemented(format!(
-                "constructor handler for {}",
-                class
-            )))
-        }
-        StateType::Chain(ct) => Err(TCError::not_implemented(format!("GET {}", ct))),
-        StateType::Map => {
-            let value = Tuple::<(Id, Value)>::try_cast_from(value, |v| {
-                TCError::bad_request("invalid Map", v)
-            })?;
-
-            let map = value
-                .into_iter()
-                .map(|(id, value)| (id, State::from(value)))
-                .collect();
-
-            Ok(State::Map(map))
-        }
-        StateType::Object(ot) => Err(TCError::not_implemented(format!("GET {}", ot))),
-        StateType::Scalar(class) => {
-            let err = format!("Cannot cast into {} from {}", class, value);
-            State::Scalar(Scalar::Value(value))
-                .into_type(StateType::Scalar(class))
-                .ok_or_else(|| TCError::unsupported(err))
-        }
-        StateType::Tuple => {
-            let value: Tuple<Value> = value.try_into()?;
-            Ok(State::Tuple(value.into_iter().map(State::from).collect()))
-        }
-    }
-}
-
-fn error_type(err_type: &Id) -> Option<ErrorType> {
-    match err_type.as_str() {
-        "bad_gateway" => Some(ErrorType::BadGateway),
-        "bad_request" => Some(ErrorType::BadRequest),
-        "conflict" => Some(ErrorType::Conflict),
-        "forbidden" => Some(ErrorType::Forbidden),
-        "internal" => Some(ErrorType::Internal),
-        "method_not_allowed" => Some(ErrorType::MethodNotAllowed),
-        "not_found" => Some(ErrorType::NotFound),
-        "not_implemented" => Some(ErrorType::NotImplemented),
-        "timeout" => Some(ErrorType::Timeout),
-        "unauthorized" => Some(ErrorType::Unauthorized),
-        _ => None,
-    }
-}
-
 async fn maybe_claim_leadership(cluster: &Cluster, txn: &Txn) -> TCResult<Txn> {
     if txn.has_owner() && !txn.has_leader(cluster.path()) {
         cluster.lead(txn.clone()).await
