@@ -26,6 +26,7 @@ mod http;
 mod route;
 
 pub mod chain;
+pub mod closure;
 pub mod cluster;
 pub mod collection;
 pub mod gateway;
@@ -33,14 +34,24 @@ pub mod kernel;
 pub mod object;
 pub mod scalar;
 pub mod state;
+pub mod stream;
 pub mod txn;
 
+const MIN_CACHE_SIZE: u64 = 5000;
+
 /// Initialize the transactional filesystem layer.
 pub async fn mount(
     workspace: PathBuf,
     data_dir: Option<PathBuf>,
     cache_size: u64,
 ) -> tc_error::TCResult<(fs::Dir, Option<fs::Dir>)> {
+    if cache_size < MIN_CACHE_SIZE {
+        return Err(error::TCError::unsupported(format!(
+            "the minimum cache size is {} bytes",
+            MIN_CACHE_SIZE
+        )));
+    }
+
     let cache = fs::Cache::new(cache_size);
 
     let workspace = fs::load(cache.clone(), workspace).await?;
