@@ -5,12 +5,15 @@ use std::fmt;
 use async_trait::async_trait;
 use destream::{de, en, EncodeMap};
 use futures::TryFutureExt;
+use safecast::TryCastFrom;
 
-use tc_error::TCResult;
+use tc_error::*;
 use tc_transact::IntoView;
+use tc_value::Value;
 use tcgeneric::{label, path_label, NativeClass, PathLabel, PathSegment, TCPathBuf};
 
 use crate::fs::Dir;
+use crate::scalar::Scalar;
 use crate::state::State;
 use crate::txn::Txn;
 
@@ -24,7 +27,7 @@ const ERR_DECODE_INSTANCE: &str = "Instance does not support direct decoding; us
 const PREFIX: PathLabel = path_label(&["state", "object"]);
 
 /// The type of a user-defined [`Object`].
-#[derive(Clone, Eq, PartialEq)]
+#[derive(Copy, Clone, Eq, PartialEq)]
 pub enum ObjectType {
     Class,
     Instance,
@@ -94,6 +97,38 @@ impl From<InstanceExt<State>> for Object {
     }
 }
 
+impl TryCastFrom<Object> for Scalar {
+    fn can_cast_from(object: &Object) -> bool {
+        match object {
+            Object::Class(class) => Self::can_cast_from(class),
+            Object::Instance(instance) => Self::can_cast_from(instance),
+        }
+    }
+
+    fn opt_cast_from(object: Object) -> Option<Self> {
+        match object {
+            Object::Class(class) => Self::opt_cast_from(class),
+            Object::Instance(instance) => Self::opt_cast_from(instance),
+        }
+    }
+}
+
+impl TryCastFrom<Object> for Value {
+    fn can_cast_from(object: &Object) -> bool {
+        match object {
+            Object::Class(class) => Self::can_cast_from(class),
+            Object::Instance(instance) => Self::can_cast_from(instance),
+        }
+    }
+
+    fn opt_cast_from(object: Object) -> Option<Self> {
+        match object {
+            Object::Class(class) => Self::opt_cast_from(class),
+            Object::Instance(instance) => Self::opt_cast_from(instance),
+        }
+    }
+}
+
 #[async_trait]
 impl de::FromStream for Object {
     type Context = ();
