@@ -7,10 +7,16 @@ use tc_transact::Transaction;
 use tc_value::Value;
 use tcgeneric::{Instance, PathSegment, TCPath};
 
-use crate::chain::{Chain, ChainInstance, Subject, SUBJECT};
+use crate::chain::{Chain, ChainInstance, ChainType, Subject, SUBJECT};
 
 use super::{DeleteHandler, GetHandler, Handler, PostHandler, Public, PutHandler, Route};
 
+impl Route for ChainType {
+    fn route<'a>(&'a self, _path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
+        None
+    }
+}
+
 impl Route for Subject {
     fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
         debug!("Chain::route {}", TCPath::from(path));
@@ -31,7 +37,10 @@ impl<'a> SubjectHandler<'a> {
 }
 
 impl<'a> Handler<'a> for SubjectHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 debug!("Subject::get {} {}", TCPath::from(self.path), key);
@@ -55,7 +64,10 @@ impl<'a> Handler<'a> for SubjectHandler<'a> {
         }))
     }
 
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key, value| {
             Box::pin(async move {
                 match self.subject {
@@ -88,7 +100,10 @@ impl<'a> Handler<'a> for SubjectHandler<'a> {
         }))
     }
 
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, params| {
             Box::pin(async move {
                 debug!("Subject::post {}", params);
@@ -108,7 +123,10 @@ impl<'a> Handler<'a> for SubjectHandler<'a> {
         }))
     }
 
-    fn delete(self: Box<Self>) -> Option<DeleteHandler<'a>> {
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 match self.subject {
@@ -133,18 +151,6 @@ impl<'a> Handler<'a> for SubjectHandler<'a> {
     }
 }
 
-impl Route for Chain {
-    fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
-        debug!("Chain::route {}", TCPath::from(path));
-
-        if path.len() == 1 && path[0].as_str() == "chain" {
-            Some(Box::new(ChainHandler::from(self)))
-        } else {
-            Some(Box::new(AppendHandler::new(self, path)))
-        }
-    }
-}
-
 struct AppendHandler<'a> {
     chain: &'a Chain,
     path: &'a [PathSegment],
@@ -157,14 +163,20 @@ impl<'a> AppendHandler<'a> {
 }
 
 impl<'a> Handler<'a> for AppendHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         match self.chain.subject().route(self.path) {
             Some(handler) => handler.get(),
             None => None,
         }
     }
 
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         match self.chain.subject().route(self.path) {
             Some(handler) => match handler.put() {
                 Some(put_handler) => Some(Box::new(|txn, key, value| {
@@ -173,7 +185,7 @@ impl<'a> Handler<'a> for AppendHandler<'a> {
 
                         let path = self.path.to_vec().into();
                         self.chain
-                            .append_put(txn.clone(), path, key.clone(), value.clone())
+                            .append_put(txn, path, key.clone(), value.clone())
                             .await?;
 
                         put_handler(txn, key, value).await
@@ -185,14 +197,20 @@ impl<'a> Handler<'a> for AppendHandler<'a> {
         }
     }
 
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         match self.chain.subject().route(self.path) {
             Some(handler) => handler.post(),
             None => None,
         }
     }
 
-    fn delete(self: Box<Self>) -> Option<DeleteHandler<'a>> {
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         match self.chain.subject().route(self.path) {
             Some(handler) => match handler.delete() {
                 Some(delete_handler) => Some(Box::new(|txn, key| {
@@ -224,7 +242,10 @@ impl<'a> From<&'a Chain> for ChainHandler<'a> {
 }
 
 impl<'a> Handler<'a> for ChainHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, key| {
             Box::pin(async move {
                 if key.is_none() {
@@ -236,3 +257,15 @@ impl<'a> Handler<'a> for ChainHandler<'a> {
         }))
     }
 }
+
+impl Route for Chain {
+    fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
+        debug!("Chain::route {}", TCPath::from(path));
+
+        if path.len() == 1 && path[0].as_str() == "chain" {
+            Some(Box::new(ChainHandler::from(self)))
+        } else {
+            Some(Box::new(AppendHandler::new(self, path)))
+        }
+    }
+}
