@@ -7,11 +7,11 @@ use safecast::{TryCastFrom, TryCastInto};
 
 use tc_error::*;
 use tc_transact::{Transact, Transaction};
+use tc_value::{Link, Value};
 use tcgeneric::{label, Id, Tuple};
 
 use crate::cluster::Cluster;
 use crate::route::*;
-use crate::scalar::{Link, Value};
 use crate::state::State;
 
 struct AuthorizeHandler<'a> {
@@ -19,7 +19,10 @@ struct AuthorizeHandler<'a> {
 }
 
 impl<'a> Handler<'a> for AuthorizeHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, scope| {
             Box::pin(async move {
                 let scope = scope
@@ -54,7 +57,7 @@ impl<'a> ClusterHandler<'a> {
                 .chain(&key)
                 .cloned()
                 .map(State::from)
-                .ok_or_else(|| TCError::not_found(key))
+                .ok_or_else(|| TCError::not_found(format!("{} member {}", self.cluster, key)))
         } else {
             let public_key = Bytes::from(self.cluster.public_key().to_vec());
             Ok(Value::from(public_key).into())
@@ -63,13 +66,19 @@ impl<'a> ClusterHandler<'a> {
 }
 
 impl<'a> Handler<'a> for ClusterHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, key| {
             Box::pin(future::ready(self.handle_get(key)))
         }))
     }
 
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key, value| {
             Box::pin(async move {
                 if key.is_some() {
@@ -84,7 +93,10 @@ impl<'a> Handler<'a> for ClusterHandler<'a> {
         }))
     }
 
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, params| {
             Box::pin(async move {
                 // TODO: authorize request using a scope
@@ -108,7 +120,10 @@ impl<'a> Handler<'a> for ClusterHandler<'a> {
         }))
     }
 
-    fn delete(self: Box<Self>) -> Option<DeleteHandler<'a>> {
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 key.expect_none()?;
@@ -136,7 +151,10 @@ struct GrantHandler<'a> {
 }
 
 impl<'a> Handler<'a> for GrantHandler<'a> {
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, mut params| {
             Box::pin(async move {
                 let scope = params.require(&label("scope").into())?;
@@ -161,7 +179,10 @@ struct InstallHandler<'a> {
 }
 
 impl<'a> Handler<'a> for InstallHandler<'a> {
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, link, scopes| {
             Box::pin(async move {
                 let link = link.try_cast_into(|v| {
@@ -191,7 +212,10 @@ struct ReplicaHandler<'a> {
 }
 
 impl<'a> Handler<'a> for ReplicaHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 key.expect_none()?;
@@ -203,7 +227,10 @@ impl<'a> Handler<'a> for ReplicaHandler<'a> {
         }))
     }
 
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key, link| {
             Box::pin(async move {
                 key.expect_none()?;
@@ -217,14 +244,17 @@ impl<'a> Handler<'a> for ReplicaHandler<'a> {
         }))
     }
 
-    fn delete(self: Box<Self>) -> Option<DeleteHandler<'a>> {
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, replicas| {
             Box::pin(async move {
                 let replicas = Tuple::<Link>::try_cast_from(replicas, |v| {
                     TCError::bad_request("expected a Link to a Cluster, not", v)
                 })?;
 
-                self.cluster.remove_replicas(&txn, &replicas).await
+                self.cluster.remove_replicas(txn, &replicas).await
             })
         }))
     }
