@@ -1,9 +1,9 @@
 use std::iter::FromIterator;
 
-use futures::{TryFutureExt, TryStreamExt};
+use futures::{future, StreamExt, TryFutureExt, TryStreamExt};
 use safecast::{Match, TryCastFrom, TryCastInto};
 
-use tc_btree::{BTreeInstance, BTreeType, Range};
+use tc_btree::{BTreeInstance, BTreeType, BTreeWrite, Range};
 use tc_error::*;
 use tc_transact::fs::Dir;
 use tc_transact::Transaction;
@@ -14,11 +14,75 @@ use crate::collection::{BTree, BTreeFile, Collection};
 use crate::route::{DeleteHandler, GetHandler, Handler, PostHandler, PutHandler, Route};
 use crate::scalar::Scalar;
 use crate::state::State;
+use crate::stream::TCStream;
+
+impl Route for BTreeType {
+    fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
+        if self == &Self::default() {
+            Static.route(path)
+        } else {
+            None
+        }
+    }
+}
+
+struct CopyHandler;
+
+impl<'a> Handler<'a> for CopyHandler {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|txn, mut params| {
+            Box::pin(async move {
+                let schema: Value = params.require(&label("schema").into())?;
+                let schema = tc_btree::RowSchema::try_cast_from(schema, |v| {
+                    TCError::bad_request("invalid BTree schema", v)
+                })?;
+
+                let source: TCStream = params.require(&label("source").into())?;
+                params.expect_empty()?;
+
+                let txn_id = *txn.id();
+
+                let file = txn
+                    .context()
+                    .create_file_tmp(*txn.id(), BTreeType::default())
+                    .await?;
+
+                let btree = BTreeFile::create(file, schema, txn_id).await?;
+
+                let keys = source.into_stream(txn.clone()).await?;
+                keys.map(|r| {
+                    r.and_then(|state| {
+                        Value::try_cast_from(state, |s| {
+                            TCError::bad_request("invalid BTree key", s)
+                        })
+                    })
+                })
+                .map(|r| {
+                    r.and_then(|value| {
+                        value.try_cast_into(|v| TCError::bad_request("invalid BTree key", v))
+                    })
+                })
+                .map_ok(|key| btree.insert(txn_id, key))
+                .try_buffer_unordered(num_cpus::get())
+                .try_fold((), |(), ()| future::ready(Ok(())))
+                .await?;
+
+                Ok(State::Collection(btree.into()))
+            })
+        }))
+    }
+}
 
 struct CreateHandler;
 
 impl<'a> Handler<'a> for CreateHandler {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, value| {
             Box::pin(async move {
                 let schema = tc_btree::RowSchema::try_cast_from(value, |v| {
@@ -39,25 +103,18 @@ impl<'a> Handler<'a> for CreateHandler {
     }
 }
 
-impl Route for BTreeType {
-    fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
-        if path.is_empty() && self == &Self::File {
-            Some(Box::new(CreateHandler))
-        } else {
-            None
-        }
-    }
-}
-
 struct BTreeHandler<'a, T> {
     btree: &'a T,
 }
 
-impl<'a, T: BTreeInstance> Handler<'a> for BTreeHandler<'a, T>
+impl<'a, T: BTreeInstance + BTreeWrite> Handler<'a> for BTreeHandler<'a, T>
 where
     BTree: From<T::Slice>,
 {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, range| {
             Box::pin(async move {
                 let range = cast_into_range(Scalar::Value(range))?;
@@ -67,7 +124,10 @@ where
         }))
     }
 
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key, value| {
             Box::pin(async move {
                 if key.is_some() {
@@ -93,7 +153,10 @@ where
         }))
     }
 
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, mut params| {
             Box::pin(async move {
                 let reverse = params.or_default(&label("reverse").into())?;
@@ -105,15 +168,14 @@ where
         }))
     }
 
-    fn delete(self: Box<Self>) -> Option<DeleteHandler<'a>> {
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, range| {
             Box::pin(async move {
                 let range = cast_into_range(Scalar::Value(range))?;
-                self.btree
-                    .clone()
-                    .slice(range, false)?
-                    .delete(*txn.id())
-                    .await
+                self.btree.clone().delete(*txn.id(), range).await
             })
         }))
     }
@@ -130,7 +192,10 @@ struct CountHandler<'a, T> {
 }
 
 impl<'a, T: BTreeInstance> Handler<'a> for CountHandler<'a, T> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 if key.is_some() {
@@ -157,7 +222,10 @@ struct FirstHandler<'a, T> {
 }
 
 impl<'a, T: BTreeInstance> Handler<'a> for FirstHandler<'a, T> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 if key.is_some() {
@@ -189,11 +257,15 @@ struct ReverseHandler<T> {
     btree: T,
 }
 
-impl<'a, T: BTreeInstance + 'a> Handler<'a> for ReverseHandler<T>
+impl<'a, T> Handler<'a> for ReverseHandler<T>
 where
+    T: BTreeInstance + 'a,
     BTree: From<<T as BTreeInstance>::Slice>,
 {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, key| {
             Box::pin(async move {
                 if key.is_some() {
@@ -216,6 +288,40 @@ impl<T> From<T> for ReverseHandler<T> {
     }
 }
 
+struct StreamHandler<T> {
+    btree: T,
+}
+
+impl<'a, T> Handler<'a> for StreamHandler<T>
+where
+    T: BTreeInstance + 'a,
+    BTree: From<T>,
+    BTree: From<T::Slice>,
+{
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|_txn, key| {
+            Box::pin(async move {
+                if key.is_none() {
+                    Ok(TCStream::from(BTree::from(self.btree)).into())
+                } else {
+                    let range = cast_into_range(Scalar::Value(key))?;
+                    let slice = self.btree.slice(range, false)?;
+                    Ok(TCStream::from(BTree::from(slice)).into())
+                }
+            })
+        }))
+    }
+}
+
+impl<T> From<T> for StreamHandler<T> {
+    fn from(btree: T) -> Self {
+        Self { btree }
+    }
+}
+
 impl Route for BTree {
     fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
         route(self, path)
@@ -229,11 +335,10 @@ impl Route for BTreeFile {
 }
 
 #[inline]
-fn route<'a, T: BTreeInstance>(
-    btree: &'a T,
-    path: &'a [PathSegment],
-) -> Option<Box<dyn Handler<'a> + 'a>>
+fn route<'a, T>(btree: &'a T, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>>
 where
+    T: BTreeInstance + BTreeWrite + 'a,
+    BTree: From<T>,
     BTree: From<T::Slice>,
 {
     if path.is_empty() {
@@ -242,6 +347,7 @@ where
         match path[0].as_str() {
             "count" => Some(Box::new(CountHandler::from(btree))),
             "first" => Some(Box::new(FirstHandler::from(btree))),
+            "keys" => Some(Box::new(StreamHandler::from(btree.clone()))),
             "reverse" => Some(Box::new(ReverseHandler::from(btree.clone()))),
             _ => None,
         }
@@ -250,6 +356,20 @@ where
     }
 }
 
+pub struct Static;
+
+impl Route for Static {
+    fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
+        if path.is_empty() {
+            Some(Box::new(CreateHandler))
+        } else if path == &["copy_from"] {
+            Some(Box::new(CopyHandler))
+        } else {
+            None
+        }
+    }
+}
+
 #[inline]
 fn cast_into_range(scalar: Scalar) -> TCResult<Range> {
     if scalar.is_none() {
