@@ -1,36 +1,205 @@
+use std::convert::TryInto;
+
 use afarray::Array;
-use futures::{Future, TryFutureExt};
+use futures::{future, Future, StreamExt, TryFutureExt, TryStreamExt};
 use log::debug;
 use safecast::{Match, TryCastFrom, TryCastInto};
 
 use tc_btree::Node;
 use tc_error::*;
 use tc_tensor::*;
-use tc_transact::fs::Dir;
+use tc_transact::fs::{CopyFrom, Dir};
 use tc_transact::Transaction;
+use tc_value::{Bound, Number, NumberClass, Range, TCString, Value, ValueType};
 use tcgeneric::{label, PathSegment, TCBoxTryFuture, Tuple};
 
-use crate::collection::{Collection, Tensor};
+use crate::collection::{Collection, DenseTensor, DenseTensorFile, SparseTensor, Tensor};
 use crate::fs;
-use crate::route::{GetHandler, PostHandler, PutHandler};
-use crate::scalar::{Bound, Number, NumberClass, Range, Value};
+use crate::route::{AttributeHandler, GetHandler, PostHandler, PutHandler};
 use crate::state::State;
+use crate::stream::TCStream;
 use crate::txn::Txn;
 
 use super::{Handler, Route};
 
+struct CastHandler<T> {
+    tensor: T,
+}
+
+impl<'a, T> Handler<'a> for CastHandler<T>
+where
+    T: TensorTransform + Send + Sync + 'a,
+    Tensor: From<T::Cast>,
+{
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|_txn, key| {
+            Box::pin(async move {
+                let dtype =
+                    ValueType::try_cast_from(key, |v| TCError::bad_request("not a NumberType", v))?;
+
+                let dtype = dtype.try_into()?;
+                self.tensor
+                    .cast_into(dtype)
+                    .map(Tensor::from)
+                    .map(State::from)
+            })
+        }))
+    }
+}
+
+impl<T> From<T> for CastHandler<T> {
+    fn from(tensor: T) -> Self {
+        Self { tensor }
+    }
+}
+
 struct ConstantHandler;
 
 impl<'a> Handler<'a> for ConstantHandler {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
-                if key.matches::<(Vec<u64>, Number)>() {
-                    let (shape, value): (Vec<u64>, Number) = key.opt_cast_into().unwrap();
-                    constant(&txn, shape, value).await
-                } else {
-                    Err(TCError::bad_request("invalid tensor schema", key))
-                }
+                let (shape, value): (Vec<u64>, Number) =
+                    key.try_cast_into(|v| TCError::bad_request("invalid Tensor schema", v))?;
+
+                constant(&txn, shape.into(), value)
+                    .map_ok(Tensor::from)
+                    .map_ok(Collection::from)
+                    .map_ok(State::from)
+                    .await
+            })
+        }))
+    }
+}
+
+struct CopyHandler;
+
+impl<'a> Handler<'a> for CopyHandler {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|txn, mut params| {
+            Box::pin(async move {
+                let source = params.require(&label("tensor").into())?;
+                params.expect_empty()?;
+
+                let copy = match source {
+                    Tensor::Dense(source) => {
+                        let file = txn
+                            .context()
+                            .create_file_tmp(*txn.id(), TensorType::Dense)
+                            .await?;
+
+                        let blocks =
+                            BlockListFile::copy_from(source.into_inner(), file, txn).await?;
+
+                        DenseTensor::from(blocks.accessor()).into()
+                    }
+                    Tensor::Sparse(source) => {
+                        let dir = txn.context().create_dir_tmp(*txn.id()).await?;
+                        let table = SparseTable::copy_from(source, dir, txn).await?;
+                        SparseTensor::from(table.accessor()).into()
+                    }
+                };
+
+                Ok(State::Collection(Collection::Tensor(copy)))
+            })
+        }))
+    }
+}
+
+struct CopyDenseHandler;
+
+impl<'a> Handler<'a> for CopyDenseHandler {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|txn, mut params| {
+            Box::pin(async move {
+                let schema: Value = params.require(&label("schema").into())?;
+                let Schema { dtype, shape } =
+                    schema.try_cast_into(|v| TCError::bad_request("invalid Tensor schema", v))?;
+
+                let source: TCStream = params.require(&label("source").into())?;
+                params.expect_empty()?;
+
+                let elements = source.into_stream(txn.clone()).await?;
+                let elements = elements.map(|r| {
+                    r.and_then(|n| {
+                        Number::try_cast_from(n, |n| {
+                            TCError::bad_request("invalid Tensor element", n)
+                        })
+                    })
+                });
+
+                let txn_id = *txn.id();
+                let file = create_file(txn).await?;
+                DenseTensorFile::from_values(file, txn_id, shape, dtype, elements)
+                    .map_ok(DenseTensor::from)
+                    .map_ok(Collection::from)
+                    .map_ok(State::Collection)
+                    .await
+            })
+        }))
+    }
+}
+
+struct CopySparseHandler;
+
+impl<'a> Handler<'a> for CopySparseHandler {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|txn, mut params| {
+            Box::pin(async move {
+                let schema: Value = params.require(&label("schema").into())?;
+                let schema: Schema =
+                    schema.try_cast_into(|v| TCError::bad_request("invalid Tensor schema", v))?;
+
+                let source: TCStream = params.require(&label("source").into())?;
+                params.expect_empty()?;
+
+                let elements = source.into_stream(txn.clone()).await?;
+
+                let txn_id = *txn.id();
+                let dir = txn.context().create_dir_tmp(txn_id).await?;
+                let tensor = SparseTensor::create(&dir, schema, txn_id).await?;
+
+                let elements = elements
+                    .map(|r| {
+                        r.and_then(|state| {
+                            Value::try_cast_from(state, |s| {
+                                TCError::bad_request("invalid sparse Tensor element", s)
+                            })
+                        })
+                    })
+                    .map(|r| {
+                        r.and_then(|row| {
+                            row.try_cast_into(|v| {
+                                TCError::bad_request(
+                                    "sparse Tensor expected a (Coord, Number) tuple, found",
+                                    v,
+                                )
+                            })
+                        })
+                    });
+
+                elements
+                    .map_ok(|(coord, value)| tensor.write_value_at(txn_id, coord, value))
+                    .try_buffer_unordered(num_cpus::get())
+                    .try_fold((), |(), ()| future::ready(Ok(())))
+                    .await?;
+
+                Ok(Collection::Tensor(tensor.into()).into())
             })
         }))
     }
@@ -41,31 +210,133 @@ struct CreateHandler {
 }
 
 impl<'a> Handler<'a> for CreateHandler {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
-                if key.matches::<Schema>() {
-                    let Schema { shape, dtype } = key.opt_cast_into().unwrap();
+                let schema: Schema =
+                    key.try_cast_into(|v| TCError::bad_request("invalid Tensor schema", v))?;
 
-                    match self.class {
-                        TensorType::Dense => constant(&txn, shape, dtype.zero()).await,
-                        TensorType::Sparse => Err(TCError::not_implemented("create sparse tensor")),
+                match self.class {
+                    TensorType::Dense => {
+                        constant(&txn, schema.shape, schema.dtype.zero())
+                            .map_ok(Tensor::from)
+                            .map_ok(Collection::Tensor)
+                            .map_ok(State::Collection)
+                            .await
+                    }
+                    TensorType::Sparse => {
+                        let txn_id = *txn.id();
+                        let dir = txn.context().create_dir_tmp(txn_id).await?;
+
+                        SparseTensor::create(&dir, schema, txn_id)
+                            .map_ok(Tensor::from)
+                            .map_ok(Collection::Tensor)
+                            .map_ok(State::Collection)
+                            .await
                     }
-                } else {
-                    Err(TCError::bad_request(
-                        "invalid schema for constant tensor",
-                        key,
-                    ))
                 }
             })
         }))
     }
 }
 
+struct EinsumHandler;
+
+impl<'a> Handler<'a> for EinsumHandler {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|_txn, mut params| {
+            Box::pin(async move {
+                let format: TCString = params.require(&label("format").into())?;
+                let tensors: Vec<Tensor> = params.require(&label("tensors").into())?;
+                einsum(&format, tensors)
+                    .map(Collection::from)
+                    .map(State::from)
+            })
+        }))
+    }
+}
+
+struct ElementsHandler<T> {
+    tensor: T,
+}
+
+impl<T> ElementsHandler<T> {
+    fn new(tensor: T) -> Self {
+        Self { tensor }
+    }
+}
+
+impl<'a, T> Handler<'a> for ElementsHandler<T>
+where
+    T: TensorAccess + TensorTransform + Send + Sync + 'a,
+    Tensor: From<T> + From<T::Slice>,
+{
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|_txn, key| {
+            Box::pin(async move {
+                let tensor = if key.is_none() {
+                    Tensor::from(self.tensor)
+                } else {
+                    let bounds = cast_bounds(self.tensor.shape(), key)?;
+                    let slice = self.tensor.slice(bounds)?;
+                    Tensor::from(slice)
+                };
+
+                Ok(TCStream::from(Collection::Tensor(tensor)).into())
+            })
+        }))
+    }
+}
+
+struct ExpandHandler<T> {
+    tensor: T,
+}
+
+impl<'a, T> Handler<'a> for ExpandHandler<T>
+where
+    T: TensorTransform + Send + 'a,
+    Tensor: From<T::Expand>,
+{
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|_txn, key| {
+            Box::pin(async move {
+                let axis = key.try_cast_into(|v| TCError::bad_request("invalid tensor axis", v))?;
+
+                self.tensor
+                    .expand_dims(axis)
+                    .map(Tensor::from)
+                    .map(Collection::from)
+                    .map(State::from)
+            })
+        }))
+    }
+}
+
+impl<T> From<T> for ExpandHandler<T> {
+    fn from(tensor: T) -> Self {
+        Self { tensor }
+    }
+}
+
 struct RangeHandler;
 
 impl<'a> Handler<'a> for RangeHandler {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 if key.matches::<(Vec<u64>, Number, Number)>() {
@@ -87,18 +358,65 @@ impl<'a> Handler<'a> for RangeHandler {
     }
 }
 
+struct TransposeHandler<T> {
+    tensor: T,
+}
+
+impl<'a, T> Handler<'a> for TransposeHandler<T>
+where
+    T: TensorTransform + Send + 'a,
+    Tensor: From<T::Transpose>,
+{
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        Some(Box::new(|_txn, key| {
+            Box::pin(async move {
+                let transpose = if key.is_none() {
+                    self.tensor.transpose(None)
+                } else {
+                    let permutation = key.try_cast_into(|v| {
+                        TCError::bad_request("invalid permutation for transpose", v)
+                    })?;
+
+                    self.tensor.transpose(Some(permutation))
+                };
+
+                transpose
+                    .map(Tensor::from)
+                    .map(Collection::from)
+                    .map(State::from)
+            })
+        }))
+    }
+}
+
+impl<T> From<T> for TransposeHandler<T> {
+    fn from(tensor: T) -> Self {
+        Self { tensor }
+    }
+}
+
 impl Route for TensorType {
     fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
         if path.is_empty() {
-            Some(Box::new(CreateHandler { class: *self }))
-        } else if path.len() == 1 && self == &Self::Dense {
-            match path[0].as_str() {
+            return Some(Box::new(CreateHandler { class: *self }));
+        } else if path.len() != 1 {
+            return None;
+        }
+
+        match self {
+            Self::Dense => match path[0].as_str() {
+                "copy_from" => Some(Box::new(CopyDenseHandler)),
                 "constant" => Some(Box::new(ConstantHandler)),
                 "range" => Some(Box::new(RangeHandler)),
                 _ => None,
-            }
-        } else {
-            None
+            },
+            Self::Sparse => match path[0].as_str() {
+                "copy_from" => Some(Box::new(CopySparseHandler)),
+                _ => None,
+            },
         }
     }
 }
@@ -121,7 +439,10 @@ impl DualHandler {
 }
 
 impl<'a> Handler<'a> for DualHandler {
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, mut params| {
             Box::pin(async move {
                 let l = self.tensor;
@@ -163,19 +484,23 @@ impl<'a, T: TensorReduce<fs::Dir>> ReduceHandler<'a, T> {
 
 impl<'a, T: TensorReduce<fs::Dir> + Clone + Sync> Handler<'a> for ReduceHandler<'a, T>
 where
-    Collection: From<<T as TensorReduce<fs::Dir>>::Reduce>,
+    Tensor: From<<T as TensorReduce<fs::Dir>>::Reduce>,
 {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
                 if key.is_none() {
-                    (self.reduce_all)(self.tensor, txn)
+                    (self.reduce_all)(self.tensor, txn.clone())
                         .map_ok(Value::from)
                         .map_ok(State::from)
                         .await
                 } else {
                     let axis = key.try_cast_into(|v| TCError::bad_request("invalid axis", v))?;
                     (self.reduce)(self.tensor.clone(), axis)
+                        .map(Tensor::from)
                         .map(Collection::from)
                         .map(State::from)
                 }
@@ -198,23 +523,29 @@ where
         + Send
         + Sync,
     <T as TensorTransform>::Slice: TensorAccess + Send,
-    Collection: From<T>,
-    Collection: From<<T as TensorTransform>::Slice>,
+    Tensor: From<<T as TensorTransform>::Slice>,
 {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, key| {
             Box::pin(async move {
                 debug!("GET Tensor: {}", key);
                 let bounds = cast_bounds(self.tensor.shape(), key)?;
                 self.tensor
                     .slice(bounds)
+                    .map(Tensor::from)
                     .map(Collection::from)
                     .map(State::from)
             })
         }))
     }
 
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(move |txn, key, value| {
             debug!("PUT Tensor: {} <- {}", key, value);
             Box::pin(write(self.tensor, txn, key, value))
@@ -240,7 +571,10 @@ impl UnaryHandler {
 }
 
 impl<'a> Handler<'a> for UnaryHandler {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, key| {
             Box::pin(async move {
                 let tensor = if key.is_none() {
@@ -271,9 +605,14 @@ impl<'a, F> Handler<'a> for UnaryHandlerAsync<F>
 where
     F: Future<Output = TCResult<bool>> + Send + 'a,
 {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|txn, key| {
             Box::pin(async move {
+                let txn = txn.clone();
+
                 if key.is_none() {
                     (self.op)(self.tensor, txn).map_ok(State::from).await
                 } else {
@@ -286,17 +625,13 @@ where
     }
 }
 
-impl<B: DenseAccess<fs::File<Array>, fs::File<Node>, fs::Dir, Txn>> Route
-    for DenseTensor<fs::File<Array>, fs::File<Node>, fs::Dir, Txn, B>
-{
+impl<B: DenseWrite<fs::File<Array>, fs::File<Node>, fs::Dir, Txn>> Route for DenseTensor<B> {
     fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
         route(self, path)
     }
 }
 
-impl<A: SparseAccess<fs::File<Array>, fs::File<Node>, fs::Dir, Txn>> Route
-    for SparseTensor<fs::File<Array>, fs::File<Node>, fs::Dir, Txn, A>
-{
+impl<A: SparseWrite<fs::File<Array>, fs::File<Node>, fs::Dir, Txn>> Route for SparseTensor<A> {
     fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
         route(self, path)
     }
@@ -322,60 +657,89 @@ where
         + Clone
         + Send
         + Sync,
-    Tensor: From<T>,
-    <T as TensorTransform>::Slice: TensorAccess + Send,
     Collection: From<T>,
-    Collection: From<<T as TensorReduce<fs::Dir>>::Reduce>,
-    Collection: From<<T as TensorTransform>::Slice>,
+    Tensor: From<T>,
+    <T as TensorTransform>::Slice: TensorAccess + Send + 'a,
+    Tensor: From<<T as TensorReduce<fs::Dir>>::Reduce>,
+    Tensor: From<<T as TensorTransform>::Cast>,
+    Tensor: From<<T as TensorTransform>::Expand>,
+    Tensor: From<<T as TensorTransform>::Slice>,
+    Tensor: From<<T as TensorTransform>::Transpose>,
 {
     if path.is_empty() {
         Some(Box::new(TensorHandler::from(tensor.clone())))
     } else if path.len() == 1 {
-        let cloned = tensor.clone();
+        match path[0].as_str() {
+            // attributes
+            "shape" => {
+                return Some(Box::new(AttributeHandler::from(
+                    tensor
+                        .shape()
+                        .iter()
+                        .map(|dim| Number::from(*dim))
+                        .collect::<Tuple<Value>>(),
+                )))
+            }
+
+            // reduce ops (which require borrowing)
+            "product" => {
+                return Some(Box::new(ReduceHandler::new(
+                    tensor,
+                    TensorReduce::product,
+                    TensorReduce::product_all,
+                )))
+            }
+            "sum" => {
+                return Some(Box::new(ReduceHandler::new(
+                    tensor,
+                    TensorReduce::sum,
+                    TensorReduce::sum_all,
+                )))
+            }
+            _ => {}
+        };
+
+        let tensor = tensor.clone();
 
         match path[0].as_str() {
+            // to stream
+            "elements" => Some(Box::new(ElementsHandler::new(tensor))),
+
             // boolean ops
-            "and" => Some(Box::new(DualHandler::new(cloned, TensorBoolean::and))),
-            "or" => Some(Box::new(DualHandler::new(cloned, TensorBoolean::or))),
-            "xor" => Some(Box::new(DualHandler::new(cloned, TensorBoolean::xor))),
+            "and" => Some(Box::new(DualHandler::new(tensor, TensorBoolean::and))),
+            "or" => Some(Box::new(DualHandler::new(tensor, TensorBoolean::or))),
+            "xor" => Some(Box::new(DualHandler::new(tensor, TensorBoolean::xor))),
 
             // comparison ops
-            "eq" => Some(Box::new(DualHandler::new(cloned, TensorCompare::eq))),
-            "gt" => Some(Box::new(DualHandler::new(cloned, TensorCompare::gt))),
-            "gte" => Some(Box::new(DualHandler::new(cloned, TensorCompare::gte))),
-            "lt" => Some(Box::new(DualHandler::new(cloned, TensorCompare::lt))),
-            "lte" => Some(Box::new(DualHandler::new(cloned, TensorCompare::lte))),
-            "ne" => Some(Box::new(DualHandler::new(cloned, TensorCompare::ne))),
+            "eq" => Some(Box::new(DualHandler::new(tensor, TensorCompare::eq))),
+            "gt" => Some(Box::new(DualHandler::new(tensor, TensorCompare::gt))),
+            "gte" => Some(Box::new(DualHandler::new(tensor, TensorCompare::gte))),
+            "lt" => Some(Box::new(DualHandler::new(tensor, TensorCompare::lt))),
+            "lte" => Some(Box::new(DualHandler::new(tensor, TensorCompare::lte))),
+            "ne" => Some(Box::new(DualHandler::new(tensor, TensorCompare::ne))),
 
             // unary ops
-            "abs" => Some(Box::new(UnaryHandler::new(cloned.into(), TensorUnary::abs))),
+            "abs" => Some(Box::new(UnaryHandler::new(tensor.into(), TensorUnary::abs))),
             "all" => Some(Box::new(UnaryHandlerAsync::new(
-                cloned.into(),
+                tensor.into(),
                 TensorUnary::all,
             ))),
             "any" => Some(Box::new(UnaryHandlerAsync::new(
-                cloned.into(),
+                tensor.into(),
                 TensorUnary::any,
             ))),
-            "not" => Some(Box::new(UnaryHandler::new(cloned.into(), TensorUnary::not))),
+            "not" => Some(Box::new(UnaryHandler::new(tensor.into(), TensorUnary::not))),
 
             // basic math
-            "add" => Some(Box::new(DualHandler::new(cloned, TensorMath::add))),
-            "div" => Some(Box::new(DualHandler::new(cloned, TensorMath::div))),
-            "mul" => Some(Box::new(DualHandler::new(cloned, TensorMath::mul))),
-            "sub" => Some(Box::new(DualHandler::new(cloned, TensorMath::sub))),
+            "add" => Some(Box::new(DualHandler::new(tensor, TensorMath::add))),
+            "div" => Some(Box::new(DualHandler::new(tensor, TensorMath::div))),
+            "mul" => Some(Box::new(DualHandler::new(tensor, TensorMath::mul))),
+            "sub" => Some(Box::new(DualHandler::new(tensor, TensorMath::sub))),
 
-            // reduce ops
-            "product" => Some(Box::new(ReduceHandler::new(
-                tensor,
-                TensorReduce::product,
-                TensorReduce::product_all,
-            ))),
-            "sum" => Some(Box::new(ReduceHandler::new(
-                tensor,
-                TensorReduce::sum,
-                TensorReduce::sum_all,
-            ))),
+            // transforms
+            "cast" => Some(Box::new(CastHandler::from(tensor))),
+            "expand_dims" => Some(Box::new(ExpandHandler::from(tensor))),
+            "transpose" => Some(Box::new(TransposeHandler::from(tensor))),
 
             _ => None,
         }
@@ -384,20 +748,34 @@ where
     }
 }
 
-async fn constant<S>(txn: &Txn, shape: S, value: Number) -> TCResult<State>
-where
-    Shape: From<S>,
-{
-    let file = create_file(txn).await?;
+pub struct Static;
 
-    DenseTensor::constant(file, *txn.id(), shape, value)
-        .map_ok(Tensor::from)
-        .map_ok(Collection::from)
-        .map_ok(State::from)
-        .await
+impl Route for Static {
+    fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
+        if path.is_empty() {
+            return None;
+        }
+
+        match path[0].as_str() {
+            "dense" => TensorType::Dense.route(&path[1..]),
+            "sparse" => TensorType::Sparse.route(&path[1..]),
+            "copy_from" if path.len() == 1 => Some(Box::new(CopyHandler)),
+            "einsum" if path.len() == 1 => Some(Box::new(EinsumHandler)),
+            _ => None,
+        }
+    }
 }
 
-async fn write<T>(tensor: T, txn: Txn, key: Value, value: State) -> TCResult<()>
+async fn constant(
+    txn: &Txn,
+    shape: Shape,
+    value: Number,
+) -> TCResult<DenseTensor<DenseTensorFile>> {
+    let file = create_file(txn).await?;
+    DenseTensor::constant(file, *txn.id(), shape, value).await
+}
+
+async fn write<T>(tensor: T, txn: &Txn, key: Value, value: State) -> TCResult<()>
 where
     T: TensorAccess
         + TensorIO<fs::Dir, Txn = Txn>
@@ -410,7 +788,9 @@ where
     let bounds = cast_bounds(tensor.shape(), key)?;
 
     match value {
-        State::Collection(Collection::Tensor(value)) => tensor.write(txn, bounds, value).await,
+        State::Collection(Collection::Tensor(value)) => {
+            tensor.write(txn.clone(), bounds, value).await
+        }
         State::Scalar(scalar) => {
             let value =
                 scalar.try_cast_into(|v| TCError::bad_request("invalid tensor element", v))?;
@@ -527,7 +907,7 @@ pub fn cast_bounds(shape: &Shape, value: Value) -> TCResult<Bounds> {
                 axes.push(bound);
             }
 
-            Ok(Bounds { axes })
+            Ok(Bounds::from(axes))
         }
         other => Err(TCError::bad_request("invalid tensor bounds", other)),
     }
