@@ -1,14 +1,16 @@
+use std::fmt;
 use std::iter::FromIterator;
 
 use log::debug;
 
 use tc_error::*;
+use tc_value::Value;
 use tcgeneric::{Id, Instance, Map, PathSegment, TCPath, Tuple};
 
 use crate::object::InstanceExt;
 use crate::route::{DeleteHandler, GetHandler, Handler, PostHandler, PutHandler, Route};
 use crate::scalar::*;
-use crate::state::State;
+use crate::state::{State, ToState};
 use crate::txn::Txn;
 
 struct GetMethod<'a, T: Instance> {
@@ -17,8 +19,11 @@ struct GetMethod<'a, T: Instance> {
     path: &'a [PathSegment],
 }
 
-impl<'a, T: Instance + Route + 'a> GetMethod<'a, T> {
-    async fn call(self, txn: Txn, key: Value) -> TCResult<State> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> GetMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    async fn call(self, txn: &Txn, key: Value) -> TCResult<State> {
         let (key_name, op_def) = self.method;
 
         let mut context = Map::new();
@@ -28,8 +33,14 @@ impl<'a, T: Instance + Route + 'a> GetMethod<'a, T> {
     }
 }
 
-impl<'a, T: Instance + Route + 'a> Handler<'a> for GetMethod<'a, T> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> Handler<'a> for GetMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(move |txn, key| Box::pin(self.call(txn, key))))
     }
 }
@@ -40,8 +51,11 @@ struct PutMethod<'a, T: Instance> {
     path: &'a [PathSegment],
 }
 
-impl<'a, T: Instance + Route + 'a> PutMethod<'a, T> {
-    async fn call(self, txn: Txn, key: Value, value: State) -> TCResult<()> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> PutMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    async fn call(self, txn: &Txn, key: Value, value: State) -> TCResult<()> {
         let (key_name, value_name, op_def) = self.method;
 
         let mut context = Map::new();
@@ -53,8 +67,14 @@ impl<'a, T: Instance + Route + 'a> PutMethod<'a, T> {
     }
 }
 
-impl<'a, T: Instance + Route + 'a> Handler<'a> for PutMethod<'a, T> {
-    fn put(self: Box<Self>) -> Option<PutHandler<'a>> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> Handler<'a> for PutMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(move |txn, key, value| {
             Box::pin(self.call(txn, key, value))
         }))
@@ -67,14 +87,23 @@ struct PostMethod<'a, T: Instance> {
     path: &'a [PathSegment],
 }
 
-impl<'a, T: Instance + Route + 'a> PostMethod<'a, T> {
-    async fn call(self, txn: Txn, params: Map<State>) -> TCResult<State> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> PostMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    async fn call(self, txn: &Txn, params: Map<State>) -> TCResult<State> {
         call_method(txn, self.subject, self.path, params, self.method).await
     }
 }
 
-impl<'a, T: Instance + Route + 'a> Handler<'a> for PostMethod<'a, T> {
-    fn post(self: Box<Self>) -> Option<PostHandler<'a>> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> Handler<'a> for PostMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(move |txn, params| {
             Box::pin(self.call(txn, params))
         }))
@@ -87,8 +116,11 @@ struct DeleteMethod<'a, T: Instance> {
     path: &'a [PathSegment],
 }
 
-impl<'a, T: Instance + Route + 'a> DeleteMethod<'a, T> {
-    async fn call(self, txn: Txn, key: Value) -> TCResult<()> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> DeleteMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    async fn call(self, txn: &Txn, key: Value) -> TCResult<()> {
         let (key_name, op_def) = self.method;
 
         let mut context = Map::new();
@@ -99,17 +131,27 @@ impl<'a, T: Instance + Route + 'a> DeleteMethod<'a, T> {
     }
 }
 
-impl<'a, T: Instance + Route + 'a> Handler<'a> for DeleteMethod<'a, T> {
-    fn delete(self: Box<Self>) -> Option<DeleteHandler<'a>> {
+impl<'a, T: Instance + Route + fmt::Display + 'a> Handler<'a> for DeleteMethod<'a, T>
+where
+    InstanceExt<T>: ToState,
+{
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(move |txn, key| Box::pin(self.call(txn, key))))
     }
 }
 
-impl<T: Instance + Route> Route for InstanceExt<T> {
+impl<T: Instance + Route + fmt::Display> Route for InstanceExt<T>
+where
+    Self: ToState,
+{
     fn route<'a>(&'a self, path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
-        debug!("InstanceExt::route {}", TCPath::from(path));
+        debug!("{} route {}", self, TCPath::from(path));
 
         if path.is_empty() {
+            debug!("routing to parent: {}", self.parent());
             self.parent().route(path)
         } else if let Some(member) = self.proto().get(&path[0]) {
             match member {
@@ -147,15 +189,18 @@ impl<T: Instance + Route> Route for InstanceExt<T> {
     }
 }
 
-async fn call_method<T: Instance + Route>(
-    txn: Txn,
+async fn call_method<T: Instance + Route + fmt::Display>(
+    txn: &Txn,
     subject: &InstanceExt<T>,
     path: &[PathSegment],
     context: Map<State>,
     form: Vec<(Id, Scalar)>,
-) -> TCResult<State> {
+) -> TCResult<State>
+where
+    InstanceExt<T>: ToState,
+{
     debug!(
-        "call method with form {:?}",
+        "call method with form {}",
         form.iter()
             .map(|(id, s)| format!("{}: {}", id, s))
             .collect::<Vec<String>>()
@@ -163,7 +208,8 @@ async fn call_method<T: Instance + Route>(
     );
 
     if !path.is_empty() {
-        return Err(TCError::not_found(TCPath::from(path)));
+        let msg = format!("{} member {}", subject, TCPath::from(path));
+        return Err(TCError::not_found(msg));
     }
 
     let capture = if let Some((capture, _)) = form.last() {
@@ -172,7 +218,7 @@ async fn call_method<T: Instance + Route>(
         return Ok(State::default());
     };
 
-    Executor::with_context(txn, subject, context.into(), form)
+    Executor::with_context(txn, Some(subject), context.into(), form)
         .capture(capture)
         .await
 }
