@@ -1,18 +1,27 @@
 use tcgeneric::PathSegment;
 
-use crate::object::{InstanceClass, InstanceExt, Object};
+use crate::object::{InstanceClass, InstanceExt, Object, ObjectType};
 use crate::state::State;
 
 use super::{GetHandler, Handler, Route};
 
 mod instance;
 
+impl Route for ObjectType {
+    fn route<'a>(&'a self, _path: &'a [PathSegment]) -> Option<Box<dyn Handler<'a> + 'a>> {
+        None
+    }
+}
+
 struct ClassHandler<'a> {
     class: &'a InstanceClass,
 }
 
 impl<'a> Handler<'a> for ClassHandler<'a> {
-    fn get(self: Box<Self>) -> Option<GetHandler<'a>> {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
         Some(Box::new(|_txn, key| {
             Box::pin(async move {
                 let parent = State::from(key);
