@@ -3,7 +3,6 @@
 use std::collections::{BTreeMap, HashSet};
 use std::convert::TryFrom;
 use std::fmt;
-use std::iter::FromIterator;
 use std::ops::{Deref, DerefMut};
 use std::str::FromStr;
 
@@ -12,23 +11,26 @@ use bytes::Bytes;
 use destream::{de, en};
 use futures::future::TryFutureExt;
 use futures::stream::{self, StreamExt, TryStreamExt};
-use log::debug;
+use log::{debug, warn};
 use safecast::{Match, TryCastFrom, TryCastInto};
 
 use tc_error::*;
+use tc_value::{Link, Number, Range, TCString, Value, ValueType};
 use tcgeneric::*;
 
+use crate::closure::Closure;
 use crate::route::Public;
-use crate::state::State;
+use crate::state::{State, ToState};
 use crate::txn::Txn;
 
 pub use op::*;
 pub use reference::*;
-pub use tc_value::*;
+pub use tc_value as value;
 
 pub mod op;
 pub mod reference;
 
+const ERR_NO_SELF: &str = "Op context has no $self--consider using a method instead";
 const PREFIX: PathLabel = path_label(&["state", "scalar"]);
 
 /// The label of an instance in its own method context
@@ -37,6 +39,7 @@ pub const SELF: Label = label("self");
 /// The [`Class`] of a [`Scalar`].
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum ScalarType {
+    Cluster,
     Map,
     Op(OpDefType),
     Range,
@@ -56,6 +59,7 @@ impl NativeClass for ScalarType {
 
         if path.len() == 3 {
             match path[2].as_str() {
+                "cluster" => Some(Self::Cluster),
                 "map" => Some(Self::Map),
                 "range" => Some(Self::Range),
                 "tuple" => Some(Self::Tuple),
@@ -77,6 +81,7 @@ impl NativeClass for ScalarType {
         let prefix = TCPathBuf::from(PREFIX);
 
         match self {
+            Self::Cluster => prefix.append(label("cluster")),
             Self::Map => prefix.append(label("map")),
             Self::Op(odt) => odt.path(),
             Self::Range => prefix.append(label("range")),
@@ -96,6 +101,7 @@ impl From<ValueType> for ScalarType {
 impl fmt::Display for ScalarType {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
+            Self::Cluster => f.write_str("Cluster reference"),
             Self::Map => f.write_str("Map<Scalar>"),
             Self::Op(odt) => fmt::Display::fmt(odt, f),
             Self::Range => f.write_str("Range"),
@@ -106,9 +112,44 @@ impl fmt::Display for ScalarType {
     }
 }
 
+/// A reference to a `Cluster`
+#[derive(Clone, Eq, PartialEq)]
+pub struct ClusterRef(TCPathBuf);
+
+impl ClusterRef {
+    /// Consume this reference and return its `TCPathBuf`
+    pub fn into_path(self) -> TCPathBuf {
+        self.0
+    }
+
+    /// Borrow this reference's `TCPathBuf`
+    pub fn path(&self) -> &TCPathBuf {
+        &self.0
+    }
+}
+
+impl From<TCPathBuf> for ClusterRef {
+    fn from(path: TCPathBuf) -> Self {
+        Self(path)
+    }
+}
+
+impl fmt::Debug for ClusterRef {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "reference to Cluster at {:?}", self.0)
+    }
+}
+
+impl fmt::Display for ClusterRef {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "reference to Cluster at {}", self.0)
+    }
+}
+
 /// A scalar value, i.e. one which is always held in main memory and never split into blocks.
 #[derive(Clone, Eq, PartialEq)]
 pub enum Scalar {
+    Cluster(ClusterRef),
     Map(Map<Self>),
     Op(OpDef),
     Range(Range),
@@ -118,7 +159,15 @@ pub enum Scalar {
 }
 
 impl Scalar {
-    /// Return true if self is an empty tuple, default link, or `Value::None`.
+    /// Return `true` if this `Scalar` variant is a [`Map`].
+    pub fn is_map(&self) -> bool {
+        match self {
+            Self::Map(_) => true,
+            _ => false,
+        }
+    }
+
+    /// Return `true` if self is an empty tuple, default link, or `Value::None`.
     pub fn is_none(&self) -> bool {
         match self {
             Self::Map(map) => map.is_empty(),
@@ -128,7 +177,7 @@ impl Scalar {
         }
     }
 
-    /// Return true if self is a reference type which needs to be resolved.
+    /// Return `true` if this is a reference type which needs to be resolved.
     pub fn is_ref(&self) -> bool {
         match self {
             Self::Map(map) => map.values().any(Self::is_ref),
@@ -138,6 +187,15 @@ impl Scalar {
         }
     }
 
+    /// Return `true` if this `Scalar` variant is a [`Tuple`].
+    pub fn is_tuple(&self) -> bool {
+        match self {
+            Self::Tuple(_) => true,
+            Self::Value(value) => value.is_tuple(),
+            _ => false,
+        }
+    }
+
     /// Cast this `Scalar` into the specified [`ScalarType`], if possible.
     pub fn into_type(self, class: ScalarType) -> Option<Self> {
         debug!("cast into {} from {}: {}", class, self.class(), self);
@@ -152,6 +210,7 @@ impl Scalar {
         use ScalarType as ST;
 
         match class {
+            ST::Cluster => self.opt_cast_into().map(ClusterRef).map(Self::Cluster),
             ST::Map => self.opt_cast_into().map(Self::Map),
             ST::Op(odt) => match odt {
                 ODT::Get => self.opt_cast_into().map(OpDef::Get).map(Self::Op),
@@ -171,6 +230,13 @@ impl Scalar {
                     .map(Box::new)
                     .map(Scalar::Ref),
 
+                RT::Before => self
+                    .opt_cast_into()
+                    .map(Box::new)
+                    .map(TCRef::Before)
+                    .map(Box::new)
+                    .map(Scalar::Ref),
+
                 RT::Case => self
                     .opt_cast_into()
                     .map(Box::new)
@@ -199,17 +265,29 @@ impl Scalar {
                         debug!("cast into DELETE ref from {}", self);
 
                         let (subject, key): (Scalar, Scalar) = self.opt_cast_into().unwrap();
-                        let delete_ref = match subject {
+                        let delete_ref: Option<(Subject, Scalar)> = match subject {
                             Scalar::Ref(tc_ref) => match *tc_ref {
                                 TCRef::Id(id_ref) => Some((id_ref.into(), key)),
                                 TCRef::Op(OpRef::Get((subject, none))) if none.is_none() => {
                                     Some((subject, key))
                                 }
-                                _ => None,
+                                other => {
+                                    debug!("invalid DELETE subject: {}", other);
+                                    None
+                                }
                             },
-                            _ => None,
+                            other => {
+                                debug!("invalid DELETE subject: {}", other);
+                                None
+                            }
                         };
 
+                        if let Some((subject, key)) = &delete_ref {
+                            debug!("DELETE ref is {}: {}", subject, key);
+                        } else {
+                            debug!("could not cast into DELETE ref");
+                        }
+
                         delete_ref.map(OpRef::Delete).map(TCRef::Op).map(Self::from)
                     } else if self.matches::<Tuple<Scalar>>() {
                         let tuple: Tuple<Scalar> = self.opt_cast_into().unwrap();
@@ -239,6 +317,20 @@ impl Scalar {
                         None
                     }
                 }
+
+                RT::While => self
+                    .opt_cast_into()
+                    .map(Box::new)
+                    .map(TCRef::While)
+                    .map(Box::new)
+                    .map(Scalar::Ref),
+
+                RT::With => self
+                    .opt_cast_into()
+                    .map(Box::new)
+                    .map(TCRef::With)
+                    .map(Box::new)
+                    .map(Scalar::Ref),
             },
 
             ST::Value(vt) => Value::opt_cast_from(self)
@@ -266,6 +358,7 @@ impl Instance for Scalar {
     fn class(&self) -> ScalarType {
         use ScalarType as ST;
         match self {
+            Self::Cluster(_) => ST::Cluster,
             Self::Map(_) => ST::Map,
             Self::Op(op) => ST::Op(op.class()),
             Self::Range(_) => ST::Range,
@@ -305,12 +398,11 @@ impl Refer for Scalar {
         }
     }
 
-    fn is_derived_write(&self) -> bool {
+    fn is_conditional(&self) -> bool {
         match self {
-            Self::Map(map) => map.values().any(|scalar| scalar.is_derived_write()),
-            Self::Op(op_def) => op_def.is_derived_write(),
-            Self::Ref(tc_ref) => tc_ref.is_derived_write(),
-            Self::Tuple(tuple) => tuple.iter().any(|scalar| scalar.is_derived_write()),
+            Self::Map(map) => map.values().any(|scalar| scalar.is_conditional()),
+            Self::Ref(tc_ref) => tc_ref.is_conditional(),
+            Self::Tuple(tuple) => tuple.iter().any(|scalar| scalar.is_conditional()),
             _ => false,
         }
     }
@@ -320,11 +412,15 @@ impl Refer for Scalar {
             Self::Map(map) => map
                 .values()
                 .any(|scalar| scalar.is_inter_service_write(cluster_path)),
+
             Self::Op(op_def) => op_def.is_inter_service_write(cluster_path),
+
             Self::Ref(tc_ref) => tc_ref.is_inter_service_write(cluster_path),
+
             Self::Tuple(tuple) => tuple
                 .iter()
                 .any(|scalar| scalar.is_inter_service_write(cluster_path)),
+
             _ => false,
         }
     }
@@ -373,12 +469,12 @@ impl Refer for Scalar {
         }
     }
 
-    async fn resolve<'a, T: Instance + Public>(
+    async fn resolve<'a, T: ToState + Instance + Public>(
         self,
         context: &'a Scope<'a, T>,
         txn: &'a Txn,
     ) -> TCResult<State> {
-        debug!("Scalar::resolve {}", self);
+        debug!("Scalar::resolve {:?}", self);
 
         match self {
             Self::Map(map) => {
@@ -430,6 +526,12 @@ impl From<Map<Scalar>> for Scalar {
     }
 }
 
+impl From<Number> for Scalar {
+    fn from(n: Number) -> Self {
+        Self::Value(n.into())
+    }
+}
+
 impl From<OpRef> for Scalar {
     fn from(op_ref: OpRef) -> Self {
         Self::Ref(Box::new(TCRef::Op(op_ref)))
@@ -533,6 +635,16 @@ impl TryCastFrom<Scalar> for Bytes {
     }
 }
 
+impl TryCastFrom<Scalar> for Closure {
+    fn can_cast_from(scalar: &Scalar) -> bool {
+        OpDef::can_cast_from(scalar)
+    }
+
+    fn opt_cast_from(scalar: Scalar) -> Option<Self> {
+        OpDef::opt_cast_from(scalar).map(Self::from)
+    }
+}
+
 impl TryCastFrom<Scalar> for OpDef {
     fn can_cast_from(scalar: &Scalar) -> bool {
         match scalar {
@@ -663,6 +775,22 @@ impl TryCastFrom<Scalar> for Number {
     }
 }
 
+impl TryCastFrom<Scalar> for TCString {
+    fn can_cast_from(scalar: &Scalar) -> bool {
+        match scalar {
+            Scalar::Value(value) => Self::can_cast_from(value),
+            _ => false,
+        }
+    }
+
+    fn opt_cast_from(scalar: Scalar) -> Option<Self> {
+        match scalar {
+            Scalar::Value(value) => Self::opt_cast_from(value),
+            _ => None,
+        }
+    }
+}
+
 impl<T: Clone + TryCastFrom<Scalar>> TryCastFrom<Scalar> for Map<T> {
     fn can_cast_from(scalar: &Scalar) -> bool {
         match scalar {
@@ -804,6 +932,7 @@ impl TryCastFrom<Scalar> for Id {
     fn can_cast_from(scalar: &Scalar) -> bool {
         match scalar {
             Scalar::Value(value) => Self::can_cast_from(value),
+            Scalar::Ref(tc_ref) => Self::can_cast_from(&**tc_ref),
             _ => false,
         }
     }
@@ -811,6 +940,7 @@ impl TryCastFrom<Scalar> for Id {
     fn opt_cast_from(scalar: Scalar) -> Option<Self> {
         match scalar {
             Scalar::Value(value) => Self::opt_cast_from(value),
+            Scalar::Ref(tc_ref) => Self::opt_cast_from(*tc_ref),
             _ => None,
         }
     }
@@ -1035,9 +1165,7 @@ impl de::Visitor for ScalarVisitor {
             if let Ok(path) = TCPathBuf::from_str(&key) {
                 if let Some(class) = ScalarType::from_path(&path) {
                     debug!("decode instance of {}", class);
-                    if let Ok(scalar) = Self::visit_map_value(class, &mut access).await {
-                        return Ok(scalar);
-                    }
+                    return Self::visit_map_value(class, &mut access).await;
                 } else {
                     debug!("not a scalar classpath: {}", path);
                 }
@@ -1066,7 +1194,7 @@ impl de::Visitor for ScalarVisitor {
         let mut items: Vec<Scalar> = if let Some(size) = access.size_hint() {
             Vec::with_capacity(size)
         } else {
-            vec![]
+            Vec::new()
         };
 
         while let Some(value) = access.next_element(()).await? {
@@ -1089,6 +1217,7 @@ impl de::FromStream for Scalar {
 impl<'en> en::ToStream<'en> for Scalar {
     fn to_stream<E: en::Encoder<'en>>(&'en self, e: E) -> Result<E::Ok, E::Error> {
         match self {
+            Scalar::Cluster(ClusterRef(path)) => single_entry(self.class().path(), path, e),
             Scalar::Map(map) => map.to_stream(e),
             Scalar::Op(op_def) => op_def.to_stream(e),
             Scalar::Range(range) => single_entry(self.class().path(), range, e),
@@ -1104,6 +1233,7 @@ impl<'en> en::IntoStream<'en> for Scalar {
         let classpath = self.class().path();
 
         match self {
+            Scalar::Cluster(ClusterRef(path)) => single_entry(classpath, path, e),
             Scalar::Map(map) => map.into_inner().into_stream(e),
             Scalar::Op(op_def) => op_def.into_stream(e),
             Scalar::Range(range) => single_entry(classpath, range, e),
@@ -1117,6 +1247,7 @@ impl<'en> en::IntoStream<'en> for Scalar {
 impl fmt::Debug for Scalar {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
+            Scalar::Cluster(cluster) => fmt::Debug::fmt(cluster, f),
             Scalar::Map(map) => fmt::Debug::fmt(map, f),
             Scalar::Op(op) => fmt::Debug::fmt(op, f),
             Scalar::Range(range) => fmt::Debug::fmt(range, f),
@@ -1130,6 +1261,7 @@ impl fmt::Debug for Scalar {
 impl fmt::Display for Scalar {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
+            Scalar::Cluster(cluster) => fmt::Display::fmt(cluster, f),
             Scalar::Map(map) => fmt::Display::fmt(map, f),
             Scalar::Op(op) => fmt::Display::fmt(op, f),
             Scalar::Range(range) => fmt::Display::fmt(range, f),
@@ -1142,12 +1274,15 @@ impl fmt::Display for Scalar {
 
 /// The execution scope of a [`Scalar`], such as an [`OpDef`] or [`TCRef`]
 pub struct Scope<'a, T> {
-    subject: &'a T,
+    subject: Option<&'a T>,
     data: Map<State>,
 }
 
-impl<'a, T: Instance + Public> Scope<'a, T> {
-    pub fn new<S: Into<State>, I: IntoIterator<Item = (Id, S)>>(subject: &'a T, data: I) -> Self {
+impl<'a, T: ToState + Instance + Public> Scope<'a, T> {
+    pub fn new<S: Into<State>, I: IntoIterator<Item = (Id, S)>>(
+        subject: Option<&'a T>,
+        data: I,
+    ) -> Self {
         let data = data.into_iter().map(|(id, s)| (id, s.into())).collect();
 
         debug!("new execution scope: {}", data);
@@ -1155,7 +1290,7 @@ impl<'a, T: Instance + Public> Scope<'a, T> {
     }
 
     pub fn with_context<S: Into<State>, I: IntoIterator<Item = (Id, S)>>(
-        subject: &'a T,
+        subject: Option<&'a T>,
         context: Map<State>,
         data: I,
     ) -> Self {
@@ -1174,20 +1309,32 @@ impl<'a, T: Instance + Public> Scope<'a, T> {
     }
 
     pub fn resolve_id(&self, id: &Id) -> TCResult<State> {
-        if id == &SELF {
-            let subject = Subject::from((IdRef::from(Id::from(SELF)), TCPathBuf::default()));
+        debug!("resolve ID {}", id);
 
-            Ok(State::Scalar(Scalar::Ref(Box::new(TCRef::Op(OpRef::Get(
-                (subject, Scalar::default()),
-            ))))))
+        let result = if id == &SELF {
+            self.subject().map(|subject| subject.to_state())
         } else {
-            debug!("resolve ID {}", id);
-
             self.data
                 .deref()
                 .get(id)
                 .cloned()
-                .ok_or_else(|| TCError::not_found(id))
+                .ok_or_else(|| TCError::not_found(format!("state with ID {}", id)))
+        };
+
+        match result {
+            Ok(state) => {
+                debug!("{} resolved to {:?}", id, state);
+                Ok(state)
+            }
+            Err(cause) => {
+                warn!(
+                    "error resolving {} in context {}: {}",
+                    id,
+                    self.data.keys().collect::<Tuple<&Id>>(),
+                    cause
+                );
+                Err(cause)
+            }
         }
     }
 
@@ -1199,18 +1346,16 @@ impl<'a, T: Instance + Public> Scope<'a, T> {
         key: Value,
     ) -> TCResult<State> {
         if subject == &SELF {
-            debug!(
-                "{} GET {}: {}",
-                self.subject.class(),
-                TCPath::from(path),
-                key
-            );
-
-            self.subject.get(txn, path, key).await
-        } else if let Some(subject) = self.data.deref().get(subject) {
+            let subject = self.subject()?;
+            subject.get(txn, path, key).await
+        } else if let Some(subject) = self.data.deref().get(subject).cloned() {
+            let subject = subject.resolve(self, txn).await?;
             subject.get(txn, path, key).await
         } else {
-            Err(TCError::not_found(subject))
+            Err(TCError::not_found(format!(
+                "GET method subject {}",
+                subject
+            )))
         }
     }
 
@@ -1223,11 +1368,16 @@ impl<'a, T: Instance + Public> Scope<'a, T> {
         value: State,
     ) -> TCResult<()> {
         if subject == &SELF {
-            self.subject.put(txn, path, key, value).await
-        } else if let Some(subject) = self.data.deref().get(subject) {
+            let subject = self.subject()?;
+            subject.put(txn, path, key, value).await
+        } else if let Some(subject) = self.data.deref().get(subject).cloned() {
+            let subject = subject.resolve(self, txn).await?;
             subject.put(txn, path, key, value).await
         } else {
-            Err(TCError::not_found(subject))
+            Err(TCError::not_found(format!(
+                "PUT method subject {}",
+                subject
+            )))
         }
     }
 
@@ -1239,11 +1389,16 @@ impl<'a, T: Instance + Public> Scope<'a, T> {
         params: Map<State>,
     ) -> TCResult<State> {
         if subject == &SELF {
-            self.subject.post(txn, path, params).await
-        } else if let Some(subject) = self.data.deref().get(subject) {
+            let subject = self.subject()?;
+            subject.post(txn, path, params).await
+        } else if let Some(subject) = self.data.deref().get(subject).cloned() {
+            let subject = subject.resolve(self, txn).await?;
             subject.post(txn, path, params).await
         } else {
-            Err(TCError::not_found(subject))
+            Err(TCError::not_found(format!(
+                "POST method subject {}",
+                subject
+            )))
         }
     }
 
@@ -1255,18 +1410,24 @@ impl<'a, T: Instance + Public> Scope<'a, T> {
         key: Value,
     ) -> TCResult<()> {
         if subject == &SELF {
-            debug!(
-                "{} GET {}: {}",
-                self.subject.class(),
-                TCPath::from(path),
-                key
-            );
-
-            self.subject.delete(txn, path, key).await
-        } else if let Some(subject) = self.data.deref().get(subject) {
+            let subject = self.subject()?;
+            subject.delete(txn, path, key).await
+        } else if let Some(subject) = self.data.deref().get(subject).cloned() {
+            let subject = subject.resolve(self, txn).await?;
             subject.delete(txn, path, key).await
         } else {
-            Err(TCError::not_found(subject))
+            Err(TCError::not_found(format!(
+                "DELETE method subject {}",
+                subject
+            )))
+        }
+    }
+
+    fn subject(&self) -> TCResult<&T> {
+        if let Some(subject) = self.subject {
+            Ok(subject)
+        } else {
+            Err(TCError::unsupported(ERR_NO_SELF))
         }
     }
 }
@@ -1287,11 +1448,8 @@ impl<'a, T> DerefMut for Scope<'a, T> {
 
 impl<'a, T> fmt::Debug for Scope<'a, T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(
-            f,
-            "execution scope with keys {}",
-            Value::from_iter(self.keys().map(|k| Value::String(k.to_string())))
-        )
+        let ids = self.keys().collect::<Tuple<&Id>>();
+        write!(f, "execution scope with IDs {}", ids)
     }
 }
 
