@@ -1,6 +1,7 @@
 //! User-defined [`OpDef`]s
 
 use std::fmt;
+use std::iter;
 use std::str::FromStr;
 
 use async_trait::async_trait;
@@ -8,12 +9,11 @@ use destream::de::{Decoder, Error, FromStream, MapAccess, Visitor};
 use destream::en::{EncodeMap, Encoder, IntoStream, ToStream};
 use log::debug;
 
-use tc_error::*;
 use tcgeneric::*;
 
+use crate::route::{DeleteHandler, GetHandler, Handler, PostHandler, PutHandler};
 use crate::scalar::{Executor, Refer, Scalar};
 use crate::state::State;
-use crate::txn::Txn;
 
 const PREFIX: PathLabel = path_label(&["state", "scalar", "op"]);
 
@@ -96,6 +96,7 @@ pub enum OpDef {
 }
 
 impl OpDef {
+    /// Replace references to `$self` with the given `path`.
     pub fn dereference_self(self, path: &TCPathBuf) -> Self {
         match self {
             Self::Get((key_name, form)) => Self::Get((key_name, dereference_self(form, path))),
@@ -109,6 +110,7 @@ impl OpDef {
         }
     }
 
+    /// Iterate over the internal state assignments of this `OpDef`.
     pub fn form(&self) -> impl Iterator<Item = &(Id, Scalar)> {
         match self {
             Self::Get((_, form)) => form,
@@ -119,18 +121,25 @@ impl OpDef {
         .iter()
     }
 
-    pub fn is_derived_write(&self) -> bool {
-        self.form()
-            .map(|(_, provider)| provider)
-            .any(|provider| provider.is_derived_write())
+    /// Return the last assignment in this `OpDef`.
+    pub fn last(&self) -> Option<&Id> {
+        match self {
+            Self::Get((_, form)) => form.last(),
+            Self::Put((_, _, form)) => form.last(),
+            Self::Post(form) => form.last(),
+            Self::Delete((_, form)) => form.last(),
+        }
+        .map(|(id, _)| id)
     }
 
+    /// Return `true` if this `OpDef` may execute a write operation to another service.
     pub fn is_inter_service_write(&self, cluster_path: &[PathSegment]) -> bool {
         self.form()
             .map(|(_, provider)| provider)
             .any(|provider| provider.is_inter_service_write(cluster_path))
     }
 
+    /// Consume this `OpDef` and return its internal state assignments.
     pub fn into_form(self) -> Vec<(Id, Scalar)> {
         match self {
             Self::Get((_, form)) => form,
@@ -140,6 +149,7 @@ impl OpDef {
         }
     }
 
+    /// Return `true` if this is a write operation.
     pub fn is_write(&self) -> bool {
         match self {
             Self::Get(_) => false,
@@ -149,6 +159,7 @@ impl OpDef {
         }
     }
 
+    /// Replace references to the given `path` with `$self`.
     pub fn reference_self(self, path: &TCPathBuf) -> Self {
         match self {
             Self::Get((key_name, form)) => Self::Get((key_name, reference_self(form, path))),
@@ -159,25 +170,57 @@ impl OpDef {
             Self::Delete((key_name, form)) => Self::Delete((key_name, reference_self(form, path))),
         }
     }
+}
 
-    pub async fn call<S: Into<State>, I: IntoIterator<Item = (Id, State)>>(
-        op_def: Vec<(Id, S)>,
-        txn: Txn,
-        context: I,
-    ) -> TCResult<State> {
-        let capture = if let Some((id, _)) = op_def.last() {
-            id.clone()
+impl<'a> Handler<'a> for OpDef {
+    fn get<'b>(self: Box<Self>) -> Option<GetHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        if let OpDef::Get((key_name, op_def)) = *self {
+            Some(Box::new(|txn, key| {
+                Box::pin(async move {
+                    let capture = if let Some((capture, _)) = op_def.last() {
+                        capture.clone()
+                    } else {
+                        return Ok(State::default());
+                    };
+
+                    let key = State::from(key);
+                    let data = iter::once((key_name, key)).chain(
+                        op_def
+                            .into_iter()
+                            .map(|(id, provider)| (id, State::Scalar(provider))),
+                    );
+
+                    let executor: Executor<State> = Executor::new(txn, None, data);
+                    executor.capture(capture).await
+                })
+            }))
         } else {
-            return Ok(State::default());
-        };
+            None
+        }
+    }
 
-        let context = context
-            .into_iter()
-            .chain(op_def.into_iter().map(|(id, s)| (id, s.into())));
+    fn put<'b>(self: Box<Self>) -> Option<PutHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        None
+    }
 
-        Executor::new(txn, &State::default(), context)
-            .capture(capture)
-            .await
+    fn post<'b>(self: Box<Self>) -> Option<PostHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        None
+    }
+
+    fn delete<'b>(self: Box<Self>) -> Option<DeleteHandler<'a, 'b>>
+    where
+        'b: 'a,
+    {
+        None
     }
 }
 
