@@ -18,7 +18,7 @@ use tcgeneric::*;
 
 use crate::route::Public;
 use crate::scalar::{Link, Scalar, Scope, Value, SELF};
-use crate::state::State;
+use crate::state::{State, ToState};
 use crate::txn::Txn;
 
 use super::{IdRef, Refer, TCRef};
@@ -93,13 +93,6 @@ impl Subject {
         }
     }
 
-    fn is_self(&self) -> bool {
-        match self {
-            Self::Ref(id_ref, _) => id_ref.id() == &SELF,
-            _ => false,
-        }
-    }
-
     fn reference_self(self, path: &TCPathBuf) -> Self {
         match self {
             Self::Link(link) if link.path().starts_with(path) => {
@@ -271,6 +264,16 @@ impl TryFrom<Subject> for TCPathBuf {
     }
 }
 
+impl fmt::Debug for Subject {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::Link(link) => fmt::Debug::fmt(link, f),
+            Self::Ref(id_ref, path) if path.is_empty() => fmt::Debug::fmt(id_ref, f),
+            Self::Ref(id_ref, path) => write!(f, "subject: {:?} {:?}", id_ref, path),
+        }
+    }
+}
+
 impl fmt::Display for Subject {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
@@ -320,27 +323,29 @@ impl Instance for OpRef {
 impl Refer for OpRef {
     fn dereference_self(self, path: &TCPathBuf) -> Self {
         match self {
-            Self::Get((subject, key)) if subject.is_self() => {
-                Self::Get((subject.dereference_self(path), key))
+            Self::Get((subject, key)) => Self::Get((subject.dereference_self(path), key)),
+            Self::Put((subject, key, value)) => Self::Put((
+                subject.dereference_self(path),
+                key,
+                value.dereference_self(path),
+            )),
+            Self::Post((subject, params)) => {
+                if let Scalar::Map(params) = Scalar::Map(params).dereference_self(path) {
+                    Self::Post((subject.dereference_self(path), params))
+                } else {
+                    panic!("Scalar::Map::dereference_self did not return a Scalar::Map")
+                }
             }
-            Self::Put((subject, key, value)) if subject.is_self() => {
-                Self::Put((subject.dereference_self(path), key, value))
-            }
-            Self::Post((subject, params)) if subject.is_self() => {
-                Self::Post((subject.dereference_self(path), params))
-            }
-            Self::Delete((subject, key)) if subject.is_self() => {
-                Self::Delete((subject.dereference_self(path), key))
-            }
-            op_ref => op_ref,
+            Self::Delete((subject, key)) => Self::Delete((subject.dereference_self(path), key)),
         }
     }
 
-    fn is_derived_write(&self) -> bool {
+    fn is_conditional(&self) -> bool {
         match self {
-            Self::Put((Subject::Ref(id_ref, _path), _, _)) => id_ref.id() != &SELF,
-            Self::Delete((Subject::Ref(id_ref, _path), _)) => id_ref.id() != &SELF,
-            _ => false,
+            Self::Get((_, key)) => key.is_conditional(),
+            Self::Put((_, key, value)) => key.is_conditional() || value.is_conditional(),
+            Self::Post((_, params)) => params.values().any(|scalar| scalar.is_conditional()),
+            Self::Delete((_, key)) => key.is_conditional(),
         }
     }
 
@@ -360,19 +365,22 @@ impl Refer for OpRef {
 
     fn reference_self(self, path: &TCPathBuf) -> Self {
         match self {
-            Self::Get((subject, key)) if subject.is_self() => {
-                Self::Get((subject.reference_self(path), key))
-            }
-            Self::Put((subject, key, value)) if subject.is_self() => {
-                Self::Put((subject.reference_self(path), key, value))
-            }
-            Self::Post((subject, params)) if subject.is_self() => {
+            Self::Get((subject, key)) => Self::Get((subject.reference_self(path), key)),
+            Self::Put((subject, key, value)) => Self::Put((
+                subject.reference_self(path),
+                key,
+                value.reference_self(path),
+            )),
+            Self::Post((subject, params)) => {
+                let params = if let Scalar::Map(params) = Scalar::Map(params).reference_self(path) {
+                    params
+                } else {
+                    panic!("Scalar::Map::reference_self did not return a Scalar::Map")
+                };
+
                 Self::Post((subject.reference_self(path), params))
             }
-            Self::Delete((subject, key)) if subject.is_self() => {
-                Self::Delete((subject.reference_self(path), key))
-            }
-            op_ref => op_ref,
+            Self::Delete((subject, key)) => Self::Delete((subject.reference_self(path), key)),
         }
     }
 
@@ -400,26 +408,32 @@ impl Refer for OpRef {
         }
     }
 
-    async fn resolve<'a, T: Instance + Public>(
+    async fn resolve<'a, T: ToState + Instance + Public>(
         self,
         context: &'a Scope<'a, T>,
         txn: &'a Txn,
     ) -> TCResult<State> {
         debug!("OpRef::resolve {} from context {:?}", self, context);
 
-        let invalid_key = |v: &State| TCError::bad_request("invalid key", v);
+        #[inline]
+        fn invalid_key<'a, T>(subject: &'a T) -> impl FnOnce(&State) -> TCError + 'a
+        where
+            T: fmt::Display + 'a,
+        {
+            move |v| TCError::unsupported(format!("{} is not a valid key for {}", v, subject))
+        }
 
         match self {
             Self::Get((subject, key)) => match subject {
                 Subject::Link(link) => {
                     let key = key.resolve(context, txn).await?;
-                    let key = key.try_cast_into(invalid_key)?;
+                    let key = key.try_cast_into(invalid_key(&link))?;
 
                     txn.get(link, key).await
                 }
                 Subject::Ref(id_ref, path) => {
                     let key = key.resolve(context, txn).await?;
-                    let key = key.try_cast_into(invalid_key)?;
+                    let key = key.try_cast_into(invalid_key(&id_ref))?;
 
                     context.resolve_get(txn, id_ref.id(), &path, key).await
                 }
@@ -427,22 +441,24 @@ impl Refer for OpRef {
             Self::Put((subject, key, value)) => match subject {
                 Subject::Link(link) => {
                     let key = key.resolve(context, txn).await?;
-                    let key = key.try_cast_into(invalid_key)?;
+                    let key = key.try_cast_into(invalid_key(&link))?;
 
                     let value = value.resolve(context, txn).await?;
 
-                    txn.put(link, key, value).map_ok(State::from).await
+                    txn.put(link, key, value)
+                        .map_ok(|()| State::default())
+                        .await
                 }
                 Subject::Ref(id_ref, path) => {
                     let key = key.resolve(context, txn);
                     let value = value.resolve(context, txn);
                     let (key, value) = try_join!(key, value)?;
 
-                    let key = key.try_cast_into(invalid_key)?;
+                    let key = key.try_cast_into(invalid_key(&id_ref))?;
 
                     context
                         .resolve_put(txn, id_ref.id(), &path, key, value)
-                        .map_ok(State::from)
+                        .map_ok(|()| State::default())
                         .await
                 }
             },
@@ -461,17 +477,17 @@ impl Refer for OpRef {
             Self::Delete((subject, key)) => match subject {
                 Subject::Link(link) => {
                     let key = key.resolve(context, txn).await?;
-                    let key = key.try_cast_into(invalid_key)?;
+                    let key = key.try_cast_into(invalid_key(&link))?;
 
-                    txn.delete(link, key).map_ok(State::from).await
+                    txn.delete(link, key).map_ok(|()| State::default()).await
                 }
                 Subject::Ref(id_ref, path) => {
                     let key = key.resolve(context, txn).await?;
-                    let key = key.try_cast_into(invalid_key)?;
+                    let key = key.try_cast_into(invalid_key(&id_ref))?;
 
                     context
                         .resolve_delete(txn, id_ref.id(), &path, key)
-                        .map_ok(State::from)
+                        .map_ok(|()| State::default())
                         .await
                 }
             },
@@ -521,7 +537,13 @@ impl OpRefVisitor {
                 let key = tuple.pop().unwrap();
                 Ok(OpRef::Put((subject, key, value)))
             }
-            other => Err(de::Error::invalid_value(other, "OpRef parameters")),
+            other => {
+                debug!(
+                    "invalid parameters for method call on {:?}: {:?}",
+                    subject, other
+                );
+                Err(de::Error::invalid_value(other, "OpRef parameters"))
+            }
         }
     }
 }
@@ -609,10 +631,12 @@ impl fmt::Display for OpRef {
         let class = self.class();
 
         match self {
-            OpRef::Get((link, id)) => write!(f, "{} {}?key={}", class, link, id),
-            OpRef::Put((path, id, val)) => write!(f, "{} {}?key={} <- {}", class, path, id, val),
-            OpRef::Post((path, _)) => write!(f, "{} {}", class, path),
-            OpRef::Delete((link, id)) => write!(f, "{} {}?key={}", class, link, id),
+            OpRef::Get((link, key)) => write!(f, "{} {}?key={}", class, link, key),
+            OpRef::Put((path, key, value)) => {
+                write!(f, "{} {}?key={} <- {}", class, path, key, value)
+            }
+            OpRef::Post((path, params)) => write!(f, "{} {}({})", class, path, params),
+            OpRef::Delete((link, key)) => write!(f, "{} {}?key={}", class, link, key),
         }
     }
 }
