@@ -9,33 +9,51 @@ use async_trait::async_trait;
 use bytes::Bytes;
 use destream::de;
 use futures::future::TryFutureExt;
-use futures::stream::{self, StreamExt, TryStreamExt};
+use futures::stream::{StreamExt, TryStreamExt};
 use log::debug;
-use safecast::TryCastFrom;
+use safecast::{CastFrom, CastInto, TryCastFrom};
 
 use tc_error::*;
 use tc_transact::Transaction;
+use tc_value::{Link, Number, TCString, Value, ValueType};
 use tcgeneric::*;
 
 use crate::chain::*;
+use crate::closure::*;
 use crate::collection::*;
 use crate::object::{Object, ObjectType};
 use crate::route::Public;
 use crate::scalar::*;
+use crate::stream::TCStream;
 use crate::txn::Txn;
 
-mod view;
-
 pub use view::StateView;
 
+mod view;
+
+pub trait ToState {
+    fn to_state(&self) -> State;
+}
+
+impl<T: Clone> ToState for T
+where
+    State: From<T>,
+{
+    fn to_state(&self) -> State {
+        self.clone().into()
+    }
+}
+
 /// The [`Class`] of a [`State`].
-#[derive(Clone, Eq, PartialEq)]
+#[derive(Copy, Clone, Eq, PartialEq)]
 pub enum StateType {
-    Collection(CollectionType),
     Chain(ChainType),
+    Collection(CollectionType),
+    Closure,
     Map,
     Object(ObjectType),
     Scalar(ScalarType),
+    Stream,
     Tuple,
 }
 
@@ -73,14 +91,47 @@ impl NativeClass for StateType {
         match self {
             Self::Collection(ct) => ct.path(),
             Self::Chain(ct) => ct.path(),
+            Self::Closure => path_label(&["state", "closure"]).into(),
             Self::Map => path_label(&["state", "map"]).into(),
             Self::Object(ot) => ot.path(),
             Self::Scalar(st) => st.path(),
+            Self::Stream => path_label(&["state", "stream"]).into(),
             Self::Tuple => path_label(&["state", "tuple"]).into(),
         }
     }
 }
 
+// impl StateClass for StateType {
+//     type Get = State;
+//
+//     fn try_cast_from_value(self, value: Value) -> TCResult<Self::Get> {
+//         match self {
+//             Self::Chain(ct) => ct.try_cast_from_value(value).map(State::from),
+//             Self::Collection(ct) => ct.try_cast_from_value(value).map(State::from),
+//             Self::Closure => Err(TCError::bad_request("cannot cast into closure from", value)),
+//             Self::Map => {
+//                 let contents: Tuple<Value> = value.try_into()?;
+//                 let contents: Vec<(Id, Value)> = contents
+//                     .try_cast_into(|v| TCError::bad_request("cannot cast into Map from", v))?;
+//
+//                 Ok(State::Map(
+//                     contents
+//                         .into_iter()
+//                         .map(|(id, value)| (id, State::from(value)))
+//                         .collect(),
+//                 ))
+//             }
+//             Self::Object(ot) => ot.try_cast_from_value(value).map(State::Object),
+//             Self::Scalar(st) => st.try_cast_from_value(value).map(State::Scalar),
+//             Self::Stream => Err(TCError::bad_request("cannot cast into Stream from", value)),
+//             Self::Tuple => {
+//                 let contents: Tuple<Value> = value.try_into()?;
+//                 Ok(State::Tuple(contents.into_iter().collect()))
+//             }
+//         }
+//     }
+// }
+
 impl From<BTreeType> for StateType {
     fn from(btt: BTreeType) -> Self {
         CollectionType::BTree(btt).into()
@@ -118,14 +169,30 @@ impl From<ValueType> for StateType {
     }
 }
 
+impl TryFrom<StateType> for ScalarType {
+    type Error = TCError;
+
+    fn try_from(st: StateType) -> TCResult<Self> {
+        match st {
+            StateType::Scalar(st) => Ok(st),
+            other => Err(TCError::bad_request(
+                "expected Scalar class but found",
+                other,
+            )),
+        }
+    }
+}
+
 impl fmt::Display for StateType {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            Self::Collection(ct) => fmt::Display::fmt(ct, f),
             Self::Chain(ct) => fmt::Display::fmt(ct, f),
+            Self::Collection(ct) => fmt::Display::fmt(ct, f),
+            Self::Closure => f.write_str("closure"),
             Self::Map => f.write_str("Map<Id, State>"),
             Self::Object(ot) => fmt::Display::fmt(ot, f),
             Self::Scalar(st) => fmt::Display::fmt(st, f),
+            Self::Stream => f.write_str("Stream"),
             Self::Tuple => f.write_str("Tuple<State>"),
         }
     }
@@ -136,9 +203,11 @@ impl fmt::Display for StateType {
 pub enum State {
     Collection(Collection),
     Chain(Chain),
+    Closure(Closure),
     Map(Map<Self>),
     Object(Object),
     Scalar(Scalar),
+    Stream(TCStream),
     Tuple(Tuple<Self>),
 }
 
@@ -153,11 +222,29 @@ impl State {
         }
     }
 
+    /// Return true if this `State` variant is a [`Map`].
+    pub fn is_map(&self) -> bool {
+        match self {
+            Self::Scalar(scalar) => scalar.is_map(),
+            Self::Map(_) => true,
+            _ => false,
+        }
+    }
+
     /// Return false if this `State` is an empty [`Tuple`] or [`Map`], default [`Link`], or `Value::None`
     pub fn is_some(&self) -> bool {
         !self.is_none()
     }
 
+    /// Return true if this `State` variant is a [`Tuple`].
+    pub fn is_tuple(&self) -> bool {
+        match self {
+            Self::Scalar(scalar) => scalar.is_tuple(),
+            Self::Tuple(_) => true,
+            _ => false,
+        }
+    }
+
     /// Return true if this `State` is a reference that needs to be resolved.
     pub fn is_ref(&self) -> bool {
         match self {
@@ -168,20 +255,28 @@ impl State {
         }
     }
 
-    /// Cast this `State` into the given [`StateType`], if possible.
-    pub fn into_type(self, class: StateType) -> Option<Self> {
-        if self.class() == class {
-            return Some(self);
+    /// Return this `State` as a [`Map`] of [`State`]s, or an error if this is not possible.
+    pub fn try_into_map<Err: Fn(State) -> TCError>(self, err: Err) -> TCResult<Map<State>> {
+        match self {
+            State::Map(states) => Ok(states),
+            State::Scalar(Scalar::Map(states)) => Ok(states
+                .into_iter()
+                .map(|(id, scalar)| (id, State::Scalar(scalar)))
+                .collect()),
+            other => Err((err)(other)),
         }
+    }
 
-        match class {
-            StateType::Scalar(class) => {
-                debug!("cast into {} from {}", class, self);
-                Scalar::opt_cast_from(self)
-                    .and_then(|scalar| scalar.into_type(class))
-                    .map(Self::Scalar)
+    pub fn try_into_tuple<Err: Fn(State) -> TCError>(self, err: Err) -> TCResult<Tuple<State>> {
+        match self {
+            State::Tuple(tuple) => Ok(tuple),
+            State::Scalar(Scalar::Tuple(tuple)) => {
+                Ok(tuple.into_iter().map(State::Scalar).collect())
             }
-            _ => None,
+            State::Scalar(Scalar::Value(Value::Tuple(tuple))) => {
+                Ok(tuple.into_iter().map(State::from).collect())
+            }
+            other => Err((err)(other)),
         }
     }
 }
@@ -190,6 +285,7 @@ impl State {
 impl Refer for State {
     fn dereference_self(self, path: &TCPathBuf) -> Self {
         match self {
+            Self::Closure(closure) => Self::Closure(closure.dereference_self(path)),
             Self::Map(map) => {
                 let map = map
                     .into_iter()
@@ -211,20 +307,19 @@ impl Refer for State {
         }
     }
 
-    fn is_derived_write(&self) -> bool {
+    fn is_conditional(&self) -> bool {
         match self {
-            Self::Map(map) => map.values().any(|state| state.is_derived_write()),
-
-            Self::Scalar(scalar) => scalar.is_derived_write(),
-
-            Self::Tuple(tuple) => tuple.iter().any(|state| state.is_derived_write()),
-
+            Self::Map(map) => map.values().any(|state| state.is_conditional()),
+            Self::Scalar(scalar) => scalar.is_conditional(),
+            Self::Tuple(tuple) => tuple.iter().any(|state| state.is_conditional()),
             _ => false,
         }
     }
 
     fn is_inter_service_write(&self, cluster_path: &[PathSegment]) -> bool {
         match self {
+            Self::Closure(closure) => closure.is_inter_service_write(cluster_path),
+
             Self::Map(map) => map
                 .values()
                 .any(|state| state.is_inter_service_write(cluster_path)),
@@ -241,6 +336,7 @@ impl Refer for State {
 
     fn reference_self(self, path: &TCPathBuf) -> Self {
         match self {
+            Self::Closure(closure) => Self::Closure(closure.reference_self(path)),
             Self::Map(map) => {
                 let map = map
                     .into_iter()
@@ -279,7 +375,7 @@ impl Refer for State {
         }
     }
 
-    async fn resolve<'a, T: Instance + Public>(
+    async fn resolve<'a, T: ToState + Instance + Public>(
         self,
         context: &'a Scope<'a, T>,
         txn: &'a Txn,
@@ -288,7 +384,7 @@ impl Refer for State {
 
         match self {
             Self::Map(map) => {
-                let mut resolved = stream::iter(map)
+                let mut resolved = futures::stream::iter(map)
                     .map(|(id, state)| state.resolve(context, txn).map_ok(|state| (id, state)))
                     .buffer_unordered(num_cpus::get());
 
@@ -302,7 +398,7 @@ impl Refer for State {
             Self::Scalar(scalar) => scalar.resolve(context, txn).await,
             Self::Tuple(tuple) => {
                 let len = tuple.len();
-                let mut resolved = stream::iter(tuple)
+                let mut resolved = futures::stream::iter(tuple)
                     .map(|state| state.resolve(context, txn))
                     .buffered(num_cpus::get());
 
@@ -329,11 +425,13 @@ impl Instance for State {
 
     fn class(&self) -> StateType {
         match self {
-            Self::Collection(collection) => StateType::Collection(collection.class()),
             Self::Chain(chain) => StateType::Chain(chain.class()),
+            Self::Closure(_) => StateType::Closure,
+            Self::Collection(collection) => StateType::Collection(collection.class()),
             Self::Map(_) => StateType::Map,
             Self::Object(object) => StateType::Object(object.class()),
             Self::Scalar(scalar) => StateType::Scalar(scalar.class()),
+            Self::Stream(_) => StateType::Stream,
             Self::Tuple(_) => StateType::Tuple,
         }
     }
@@ -345,15 +443,21 @@ impl From<()> for State {
     }
 }
 
+impl From<BTree> for State {
+    fn from(btree: BTree) -> Self {
+        Self::Collection(btree.into())
+    }
+}
+
 impl From<Chain> for State {
     fn from(chain: Chain) -> Self {
         Self::Chain(chain)
     }
 }
 
-impl From<BTree> for State {
-    fn from(btree: BTree) -> Self {
-        Self::Collection(btree.into())
+impl From<Closure> for State {
+    fn from(closure: Closure) -> Self {
+        Self::Closure(closure)
     }
 }
 
@@ -381,12 +485,43 @@ impl From<Map<Scalar>> for State {
     }
 }
 
+impl From<Number> for State {
+    fn from(n: Number) -> Self {
+        Self::Scalar(n.into())
+    }
+}
+
+impl From<Object> for State {
+    fn from(object: Object) -> Self {
+        State::Object(object)
+    }
+}
+
+impl From<OpDef> for State {
+    fn from(op_def: OpDef) -> Self {
+        State::Scalar(Scalar::Op(op_def))
+    }
+}
+
 impl From<OpRef> for State {
     fn from(op_ref: OpRef) -> Self {
         TCRef::Op(op_ref).into()
     }
 }
 
+impl<T> From<Option<T>> for State
+where
+    State: From<T>,
+{
+    fn from(state: Option<T>) -> Self {
+        if let Some(state) = state {
+            state.into()
+        } else {
+            Value::None.into()
+        }
+    }
+}
+
 impl From<Scalar> for State {
     fn from(scalar: Scalar) -> Self {
         State::Scalar(scalar)
@@ -399,6 +534,12 @@ impl From<Table> for State {
     }
 }
 
+impl From<TCStream> for State {
+    fn from(stream: TCStream) -> Self {
+        Self::Stream(stream)
+    }
+}
+
 #[cfg(feature = "tensor")]
 impl From<Tensor> for State {
     fn from(tensor: Tensor) -> Self {
@@ -454,6 +595,63 @@ impl From<u64> for State {
     }
 }
 
+impl<T1> CastFrom<(T1,)> for State
+where
+    State: CastFrom<T1>,
+{
+    fn cast_from(value: (T1,)) -> Self {
+        State::Tuple(vec![value.0.cast_into()].into())
+    }
+}
+
+impl<T1, T2> CastFrom<(T1, T2)> for State
+where
+    State: CastFrom<T1>,
+    State: CastFrom<T2>,
+{
+    fn cast_from(value: (T1, T2)) -> Self {
+        State::Tuple(vec![value.0.cast_into(), value.1.cast_into()].into())
+    }
+}
+
+impl<T1, T2, T3> CastFrom<(T1, T2, T3)> for State
+where
+    State: CastFrom<T1>,
+    State: CastFrom<T2>,
+    State: CastFrom<T3>,
+{
+    fn cast_from(value: (T1, T2, T3)) -> Self {
+        State::Tuple(
+            vec![
+                value.0.cast_into(),
+                value.1.cast_into(),
+                value.2.cast_into(),
+            ]
+            .into(),
+        )
+    }
+}
+
+impl<T1, T2, T3, T4> CastFrom<(T1, T2, T3, T4)> for State
+where
+    State: CastFrom<T1>,
+    State: CastFrom<T2>,
+    State: CastFrom<T3>,
+    State: CastFrom<T4>,
+{
+    fn cast_from(value: (T1, T2, T3, T4)) -> Self {
+        State::Tuple(
+            vec![
+                value.0.cast_into(),
+                value.1.cast_into(),
+                value.2.cast_into(),
+                value.3.cast_into(),
+            ]
+            .into(),
+        )
+    }
+}
+
 impl TryFrom<State> for Map<State> {
     type Error = TCError;
 
@@ -508,6 +706,40 @@ impl TryCastFrom<State> for Bytes {
     }
 }
 
+impl TryCastFrom<State> for Closure {
+    fn can_cast_from(state: &State) -> bool {
+        match state {
+            State::Closure(_) => true,
+            State::Scalar(scalar) => Self::can_cast_from(scalar),
+            _ => false,
+        }
+    }
+
+    fn opt_cast_from(state: State) -> Option<Self> {
+        match state {
+            State::Closure(closure) => Some(closure),
+            State::Scalar(scalar) => Self::opt_cast_from(scalar),
+            _ => None,
+        }
+    }
+}
+
+impl TryCastFrom<State> for TCString {
+    fn can_cast_from(state: &State) -> bool {
+        match state {
+            State::Scalar(scalar) => Self::can_cast_from(scalar),
+            _ => false,
+        }
+    }
+
+    fn opt_cast_from(state: State) -> Option<Self> {
+        match state {
+            State::Scalar(scalar) => Self::opt_cast_from(scalar),
+            _ => None,
+        }
+    }
+}
+
 impl<T: Clone + TryCastFrom<State>> TryCastFrom<State> for Map<T> {
     fn can_cast_from(state: &State) -> bool {
         match state {
@@ -526,6 +758,25 @@ impl<T: Clone + TryCastFrom<State>> TryCastFrom<State> for Map<T> {
     }
 }
 
+impl TryCastFrom<State> for TCStream {
+    fn can_cast_from(state: &State) -> bool {
+        match state {
+            // State::Chain(_) => true, // TODO
+            State::Collection(_) => true,
+            State::Stream(_) => true,
+            _ => false,
+        }
+    }
+
+    fn opt_cast_from(state: State) -> Option<Self> {
+        match state {
+            State::Collection(collection) => Some(Self::from(collection)),
+            State::Stream(stream) => Some(stream),
+            _ => None,
+        }
+    }
+}
+
 impl<T: TryCastFrom<State>> TryCastFrom<State> for (T,) {
     fn can_cast_from(state: &State) -> bool {
         match state {
@@ -592,7 +843,7 @@ impl<T: TryCastFrom<State>> TryCastFrom<State> for Vec<T> {
     }
 }
 
-impl<T: Clone + TryCastFrom<State>> TryCastFrom<State> for Tuple<T> {
+impl<T: TryCastFrom<State>> TryCastFrom<State> for Tuple<T> {
     fn can_cast_from(state: &State) -> bool {
         match state {
             State::Tuple(tuple) => Vec::<T>::can_cast_from(tuple),
@@ -692,6 +943,7 @@ impl TryCastFrom<State> for Scalar {
     fn can_cast_from(state: &State) -> bool {
         match state {
             State::Map(map) => BTreeMap::<Id, Scalar>::can_cast_from(map),
+            State::Object(object) => Self::can_cast_from(object),
             State::Scalar(_) => true,
             State::Tuple(tuple) => Vec::<Scalar>::can_cast_from(tuple),
             _ => false,
@@ -704,6 +956,8 @@ impl TryCastFrom<State> for Scalar {
                 .map(Map::from)
                 .map(Scalar::Map),
 
+            State::Object(object) => Self::opt_cast_from(object),
+
             State::Scalar(scalar) => Some(scalar),
 
             State::Tuple(tuple) => Vec::<Scalar>::opt_cast_from(tuple)
@@ -751,6 +1005,7 @@ impl TryCastFrom<State> for TCPathBuf {
 impl TryCastFrom<State> for Value {
     fn can_cast_from(state: &State) -> bool {
         match state {
+            State::Object(object) => Self::can_cast_from(object),
             State::Scalar(scalar) => Self::can_cast_from(scalar),
             State::Tuple(tuple) => tuple.iter().all(Self::can_cast_from),
             _ => false,
@@ -759,6 +1014,8 @@ impl TryCastFrom<State> for Value {
 
     fn opt_cast_from(state: State) -> Option<Self> {
         match state {
+            State::Object(object) => Self::opt_cast_from(object),
+
             State::Scalar(scalar) => Self::opt_cast_from(scalar),
 
             State::Tuple(tuple) => Vec::<Value>::opt_cast_from(tuple)
@@ -789,11 +1046,13 @@ impl TryCastFrom<State> for Number {
 impl fmt::Debug for State {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            Self::Collection(collection) => fmt::Debug::fmt(collection, f),
             Self::Chain(chain) => fmt::Debug::fmt(chain, f),
+            Self::Closure(closure) => fmt::Debug::fmt(closure, f),
+            Self::Collection(collection) => fmt::Debug::fmt(collection, f),
             Self::Map(map) => fmt::Debug::fmt(map, f),
             Self::Object(object) => fmt::Debug::fmt(object, f),
             Self::Scalar(scalar) => fmt::Debug::fmt(scalar, f),
+            Self::Stream(_) => f.write_str("Stream"),
             Self::Tuple(tuple) => fmt::Debug::fmt(tuple, f),
         }
     }
@@ -802,11 +1061,13 @@ impl fmt::Debug for State {
 impl fmt::Display for State {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            Self::Collection(collection) => fmt::Display::fmt(collection, f),
             Self::Chain(chain) => fmt::Display::fmt(chain, f),
+            Self::Closure(closure) => fmt::Display::fmt(closure, f),
+            Self::Collection(collection) => fmt::Display::fmt(collection, f),
             Self::Map(map) => fmt::Display::fmt(map, f),
             Self::Object(object) => fmt::Display::fmt(object, f),
             Self::Scalar(scalar) => fmt::Display::fmt(scalar, f),
+            Self::Stream(_) => f.write_str("Stream"),
             Self::Tuple(tuple) => fmt::Display::fmt(tuple, f),
         }
     }
@@ -824,18 +1085,24 @@ impl StateVisitor {
         access: &mut A,
     ) -> Result<State, A::Error> {
         match class {
-            StateType::Collection(ct) => {
-                CollectionVisitor::new(self.txn.clone())
-                    .visit_map_value(ct, access)
-                    .map_ok(State::Collection)
-                    .await
-            }
             StateType::Chain(ct) => {
                 ChainVisitor::new(self.txn.clone())
                     .visit_map_value(ct, access)
                     .map_ok(State::Chain)
                     .await
             }
+            StateType::Closure => {
+                access
+                    .next_value(self.txn.clone())
+                    .map_ok(State::Closure)
+                    .await
+            }
+            StateType::Collection(ct) => {
+                CollectionVisitor::new(self.txn.clone())
+                    .visit_map_value(ct, access)
+                    .map_ok(State::Collection)
+                    .await
+            }
             StateType::Map => access.next_value(self.txn.clone()).await,
             StateType::Object(ot) => match ot {
                 ObjectType::Class => {
@@ -855,6 +1122,7 @@ impl StateVisitor {
                     .map_ok(State::Scalar)
                     .await
             }
+            StateType::Stream => Err(de::Error::invalid_type("Stream", "a Collection")),
             StateType::Tuple => access.next_value(self.txn.clone()).await,
         }
     }
